import torch

from ..functional.lif_refrac import LIFRefracParameters, LIFRefracState
from ..functional.lif import LIFState
from ..functional.lif_mc_refrac import lif_mc_refrac_step

import numpy as np
from typing import Optional, Tuple


class LIFMCRefracCell(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFRefracParameters = LIFRefracParameters(),
        dt: float = 0.001,
    ):
        super(LIFMCRefracCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.g_coupling = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.p = p
        self.dt = dt
        self.hidden_size = hidden_size

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFRefracState] = None
    ) -> Tuple[torch.Tensor, LIFRefracState]:
        if state is None:
            state = LIFRefracState(
                lif=LIFState(
                    z=torch.zeros(
                        input_tensor.shape[0],
                        self.hidden_size,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    v=self.p.lif.v_leak.detach()
                    * torch.ones(
                        input_tensor.shape[0],
                        self.hidden_size,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    i=torch.zeros(
                        input_tensor.shape[0],
                        self.hidden_size,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                ),
                rho=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
        return lif_mc_refrac_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            self.g_coupling,
            p=self.p,
            dt=self.dt,
        )



"""
Stateless spiking neural network components.
"""

from .coba_lif import (
    CobaLIFState,
    CobaLIFParameters,
    CobaLIFFeedForwardState,
    coba_lif_feed_forward_step,
    coba_lif_step,
)
from .correlation_sensor import (
    CorrelationSensorParameters,
    CorrelationSensorState,
    correlation_based_update,
    correlation_sensor_step,
)
from .encode import (
    constant_current_lif_encode,
    euclidean_distance,
    gaussian_rbf,
    poisson_encode,
    population_encode,
    signed_poisson_encode,
    spike_latency_encode,
    spike_latency_lif_encode,
)
from .heaviside import heaviside
from .leaky_integrator import LIParameters, LIState, li_feed_forward_step, li_step
from .lif import (
    LIFFeedForwardState,
    LIFParameters,
    LIFState,
    lif_current_encoder,
    lif_feed_forward_step,
    lif_step,
)
from .lif_adex import (
    LIFAdExFeedForwardState,
    LIFAdExParameters,
    LIFAdExState,
    lif_adex_current_encoder,
    lif_adex_feed_forward_step,
    lif_adex_step,
)
from .lif_correlation import (
    LIFCorrelationParameters,
    LIFCorrelationState,
    lif_correlation_step,
)
from .lif_ex import (
    LIFExFeedForwardState,
    LIFExParameters,
    LIFExState,
    lif_ex_current_encoder,
    lif_ex_feed_forward_step,
    lif_ex_step,
)
from .lif_mc import lif_mc_feed_forward_step, lif_mc_step
from .lif_mc_refrac import lif_mc_refrac_feed_forward_step, lif_mc_refrac_step
from .lif_refrac import (
    LIFRefracFeedForwardState,
    LIFRefracParameters,
    LIFRefracState,
    lif_refrac_feed_forward_step,
    lif_refrac_step,
)
from .logical import logical_and, logical_or, logical_xor
from .lsnn import (
    LSNNFeedForwardState,
    LSNNParameters,
    LSNNState,
    lsnn_feed_forward_step,
    lsnn_step,
)
from .regularization import regularize_step, spike_accumulator, voltage_accumulator
from .stdp_sensor import STDPSensorParameters, STDPSensorState, stdp_sensor_step
from .threshold import (
    circ_dist_fn,
    heavi_circ_fn,
    heavi_erfc_fn,
    heavi_tanh_fn,
    heavi_tent_fn,
    logistic_fn,
)

__all__ = [
    "CobaLIFState",
    "CobaLIFParameters",
    "CobaLIFFeedForwardState",
    "coba_lif_feed_forward_step",
    "coba_lif_step",
    "CorrelationSensorParameters",
    "CorrelationSensorState",
    "correlation_based_update",
    "correlation_sensor_step",
    "constant_current_lif_encode",
    "euclidean_distance",
    "gaussian_rbf",
    "poisson_encode",
    "population_encode",
    "signed_poisson_encode",
    "spike_latency_encode",
    "spike_latency_lif_encode",
    "heaviside",
    "LIParameters",
    "LIState",
    "li_feed_forward_step",
    "li_step",
    "LIFFeedForwardState",
    "LIFParameters",
    "LIFState",
    "lif_current_encoder",
    "lif_feed_forward_step",
    "lif_step",
    "LIFAdExFeedForwardState",
    "LIFAdExParameters",
    "LIFAdExState",
    "lif_adex_current_encoder",
    "lif_adex_feed_forward_step",
    "lif_adex_step",
    "LIFCorrelationParameters",
    "LIFCorrelationState",
    "lif_correlation_step",
    "LIFExFeedForwardState",
    "LIFExParameters",
    "LIFExState",
    "lif_ex_current_encoder",
    "lif_ex_feed_forward_step",
    "lif_ex_step",
    "lif_mc_feed_forward_step",
    "lif_mc_step",
    "lif_mc_refrac_feed_forward_step",
    "lif_mc_refrac_step",
    "LIFRefracFeedForwardState",
    "LIFRefracParameters",
    "LIFRefracState",
    "lif_refrac_feed_forward_step",
    "lif_refrac_step",
    "logical_and",
    "logical_or",
    "logical_xor",
    "LSNNFeedForwardState",
    "LSNNParameters",
    "LSNNState",
    "lsnn_feed_forward_step",
    "lsnn_step",
    "regularize_step",
    "spike_accumulator",
    "voltage_accumulator",
    "STDPSensorParameters",
    "STDPSensorState",
    "stdp_sensor_step",
    "circ_dist_fn",
    "heavi_circ_fn",
    "heavi_erfc_fn",
    "heavi_tanh_fn",
    "heavi_tent_fn",
    "logistic_fn",
]



"""
This module contains ``torch.nn.Module``s for regularisation operations on
spiking layers, where it can be desirable to regularise spikes,
membrane parameters, or other properties over time.
"""
import torch

from typing import Any

from norse.torch.functional.regularization import (
    regularize_step,
    spike_accumulator,
    Accumulator,
)


class RegularizationCell(torch.nn.Module):
    """
    A regularisation cell that accumulates some state (for instance number of spikes)
    for each forward step, which can later be applied to a loss term.

    Example:
        >>> import torch
        >>> from norse.torch.module import lif, regularization
        >>> cell = lif.LIFCell(2, 4) # 2 -> 4
        >>> r = regularization.RegularizationCell() # Defaults to spike counting
        >>> data = torch.ones(5, 2)  # Batch size of 5
        >>> z, s = cell(data)
        >>> z, regularization_term = r(z, s)
        >>> ...
        >>> loss = ... + 1e-3 * regularization_term

    Parameters:
        accumulator (Accumulator):
            The accumulator that aggregates some data (such as spikes) that can later
            be included in an error term.
        state (Optional[T]): The regularization state to be aggregated to of any type T. Defaults to None.
    """

    def __init__(self, accumulator: Accumulator = spike_accumulator, state: Any = None):
        super(RegularizationCell, self).__init__()
        self.accumulator = accumulator
        self.state = state

    def forward(self, z, s):
        _, state = regularize_step(z, s, self.accumulator, self.state)
        self.state = state
        return z, self.state



from typing import Any, Optional, Tuple

import numpy as np
import torch

from ..functional.lif import (
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
    lif_step,
    lif_feed_forward_step,
)


class LIFCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a
    LIF neuron-model with recurrence.
    More specifically it implements one integration step
    of the following ODE

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\
            \\dot{i} &= -1/\\tau_{\\text{syn}} i
        \\end{align*}

    together with the jump condition

    .. math::
        z = \\Theta(v - v_{\\text{th}})

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\
            i &= i + w_{\\text{rec}} z_{\\text{rec}}
        \\end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the
    recurrent and input spikes respectively.

    Parameters:
        input_size (int): Size of the input. Also known as the number of input features.
        hidden_size (int): Size of the hidden state. Also known as the number of input features.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = LIFCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> output, s0 = lif(input)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFParameters = LIFParameters(),
        dt: float = 0.001,
    ):
        super(LIFCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) * np.sqrt(2 / hidden_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"{self.input_size}, {self.hidden_size}, p={self.p}, dt={self.dt}"
        return s

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFState] = None
    ) -> Tuple[torch.Tensor, LIFState]:
        if state is None:
            state = LIFState(
                z=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return lif_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFLayer(torch.nn.Module):
    """
    A neuron layer that wraps a recurrent LIFCell in time such
    that the layer keeps track of temporal sequences of spikes.
    After application, the layer returns a tuple containing
      (spikes from all timesteps, state from the last timestep).

    Example:
        >>> data = torch.zeros(10, 5, 2) # 10 timesteps, 5 batches, 2 neurons
        >>> l = LIFLayer(2, 4)
        >>> l(data) # Returns tuple of (Tensor(10, 5, 4), LIFState)
    """

    def __init__(self, *cell_args, **kw_args):
        super(LIFLayer, self).__init__()
        self.cell = LIFCell(*cell_args, **kw_args)

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFState] = None
    ) -> Tuple[torch.Tensor, LIFState]:
        inputs = input_tensor.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for _, input_step in enumerate(inputs):
            out, state = self.cell(input_step, state)
            outputs += [out]
        # pytype: disable=bad-return-type
        return torch.stack(outputs), state
        # pytype: enable=bad-return-type


class LIFFeedForwardCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF neuron.
    It takes as input the input current as generated by an arbitrary torch
    module or function. More specifically it implements one integration step
    of the following ODE

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\\\
            \\dot{i} &= -1/\\tau_{\\text{syn}} i
        \\end{align*}

    together with the jump condition

    .. math::
        z = \\Theta(v - v_{\\text{th}})

    and transition equations

    .. math::
        i = i + i_{\\text{in}}

    where :math:`i_{\\text{in}}` is meant to be the result of applying
    an arbitrary pytorch module (such as a convolution) to input spikes.

    Parameters:
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = LIFFeedForwardCell(20, 30)
        >>> data = torch.randn(batch_size, 20)
        >>> output, s0 = lif(data)
    """

    def __init__(self, p: LIFParameters = LIFParameters(), dt: float = 0.001):
        super(LIFFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"p={self.p}, dt={self.dt}"
        return s

    def forward(
        self, x: torch.Tensor, state: Optional[LIFFeedForwardState] = None
    ) -> Tuple[torch.Tensor, LIFFeedForwardState]:
        if state is None:
            state = LIFFeedForwardState(
                v=self.p.v_leak.detach(),
                i=torch.zeros(*x.shape, device=x.device, dtype=x.dtype),
            )
            state.v.requires_grad = True
        return lif_feed_forward_step(x, state, p=self.p, dt=self.dt)


class LIFFeedForwardLayer(torch.nn.Module):
    """
    A neuron layer that wraps a recurrent LIFCell in time such
    that the layer keeps track of temporal sequences of spikes.
    After application, the layer returns a tuple containing
      (spikes from all timesteps, state from the last timestep).

    Example:
        >>> data = torch.zeros(10, 5, 2) # 10 timesteps, 5 batches, 2 neurons
        >>> l = LIFFeedForwardLayer()
        >>> l(data) # Returns tuple of (Tensor(10, 5, 4), LIFState)

    Arguments:
        cell_args (Any): Arguments to pass on to the LIFCell
        kw_args (Any): Key-value arguments po pass on to the LIFCell
    """

    def __init__(self, *cell_args: Any, **kw_args: Any):
        super(LIFFeedForwardLayer, self).__init__()
        self.cell = LIFFeedForwardCell(*cell_args, **kw_args)

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFFeedForwardState] = None
    ) -> Tuple[torch.Tensor, LIFFeedForwardState]:
        inputs = input_tensor.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for _, input_step in enumerate(inputs):
            out, state = self.cell(input_step, state)
            outputs += [out]
        # pytype: disable=bad-return-type
        return torch.stack(outputs), state
        # pytype: enable=bad-return-type



"""
A module for lifting neuron activation functions in time.
Simlar to the :module:`.lift`_ module.
"""

import torch


class _Lifted:
    """
    Helper class for the :func:`lift`_ function to allow for pickling.
    Used in distributed execution, like PyTorch Lightning
    """

    def __init__(self, activation, p=None):
        self.activation = activation
        self.p = p

    def __call__(self, input_tensor, **kwargs):
        if self.p is not None and "p" not in kwargs:
            kwargs["p"] = self.p

        state = kwargs.get("state", None)
        kwargs.pop("state", None)
        outputs = []
        for i in input_tensor:
            if state is not None:
                out, state = self.activation(i, state=state, **kwargs)
            else:
                out, state = self.activation(i, **kwargs)
            outputs.append(out)
        return torch.stack(outputs), state


def lift(activation, p=None):
    """
    Creates a lifted version of the given activation function which
    applies the activation function in the temporal domain. The returned
    callable can be applied later as if it was a regular activation function,
    but the input is now assumed to be a tensor whose first dimension is time.

    Parameters:
        activation (Callable[[torch.Tensor, Any, Any], Tuple[torch.Tensor, Any]]):
            The activation function that takes an input tensor, an optional state, and an
            optional parameter object and returns a tuple of (spiking output, neuron state).
            The returned spiking output includes the time domain.
        p (Any): An optional parameter object to hand to the activation function.

    Returns:
        A :class:`.Callable`_ that, when applied, evaluates the activation function N times,
        where N is the size of the outer (temporal) dimension. The application will provide
        a tensor of shape (time, ...).
    """

    return _Lifted(activation, p)



from typing import Tuple, Union


import torch


class Lift(torch.nn.Module):
    """Lift applies a given torch.nn.Module over
       a temporal sequence. In other words this module
       applies the given torch.nn.Module N times, where N
       is the outer dimension in the provided tensor.

    Parameters:
        module: Module to apply

    Examples:

        >>> batch_size = 16
        >>> seq_length = 1000
        >>> in_channels = 64
        >>> out_channels = 32
        >>> conv2d = Lift(torch.nn.Conv2d(in_channels, out_channels, 5, 1))
        >>> data = torch.randn(seq_length, batch_size, 20, 30)
        >>> output = conv2d(data)


        >>> data = torch.randn(seq_length, batch_size, in_channels, 20, 30)
        >>> module = torch.nn.Sequential(
        >>>     Lift(torch.nn.Conv2d(in_channels, out_channels, 5, 1)),
        >>>     LIFFeedForwardLayer(),
        >>> )
        >>> output, _ = module(data)
    """

    def __init__(self, module: torch.nn.Module):
        super(Lift, self).__init__()
        self.lifted_module = module

    def forward(
        self, x: Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]
    ) -> torch.Tensor:
        """Apply the module over the input along the 0-th (time) dimension
        and accumulate the outputs in an output tensor.

        Parameters:
            x : Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]

        Note:
            If the input is a tuple of two tensors, the second tuple entry will be ignored.
        """
        if isinstance(x, tuple):
            x, _ = x

        T = x.shape[0]
        outputs = []

        for ts in range(T):
            out = self.lifted_module(x[ts])
            outputs += [out]

        return torch.stack(outputs)



# adapted from https://github.com/pytorch/vision/blob/master/torchvision/models/mobilenet.py
# licensed under BSD 3-Clause License, see LICENSE.torchvision for license details
from torch import nn
from torch.hub import load_state_dict_from_url
from norse.torch.module.lif import LIFFeedForwardLayer
from norse.torch.module.lift import Lift

__all__ = ["MobileNetV2", "mobilenet_v2"]


model_urls = {
    "mobilenet_v2": "https://download.pytorch.org/models/mobilenet_v2-b0353104.pth",
}


def _make_divisible(v, divisor):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:

    :return:
    """
    min_value = divisor

    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvBNReLU(nn.Sequential):
    def __init__(
        self, in_planes, out_planes, norm_layer, kernel_size=3, stride=1, groups=1
    ):
        padding = (kernel_size - 1) // 2

        super(ConvBNReLU, self).__init__(
            Lift(
                nn.Conv2d(
                    in_planes,
                    out_planes,
                    kernel_size,
                    stride,
                    padding,
                    groups=groups,
                    bias=False,
                )
            ),
            Lift(norm_layer(out_planes)),
            LIFFeedForwardLayer(),
        )


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio, norm_layer):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(
                ConvBNReLU(inp, hidden_dim, kernel_size=1, norm_layer=norm_layer)
            )
        layers.extend(
            [
                # dw
                ConvBNReLU(
                    hidden_dim,
                    hidden_dim,
                    stride=stride,
                    groups=hidden_dim,
                    norm_layer=norm_layer,
                ),
                # pw-linear
                Lift(nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False)),
                Lift(norm_layer(oup)),
            ]
        )
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(
        self,
        num_classes=1000,
        width_mult=1.0,
        inverted_residual_setting=None,
        round_nearest=8,
        block=None,
        norm_layer=None,
    ):
        """
        MobileNet V2 main class
        Args:
            num_classes (int): Number of classes
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for mobilenet
            norm_layer: Module specifying the normalization layer to use
        """
        super(MobileNetV2, self).__init__()

        if block is None:
            block = InvertedResidual

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        input_channel = 32
        last_channel = 1280

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        # only check the first element, assuming user knows t,c,n,s are required
        if (
            len(inverted_residual_setting) == 0
            or len(inverted_residual_setting[0]) != 4
        ):
            raise ValueError(
                "inverted_residual_setting should be non-empty "
                "or a 4-element list, got {}".format(inverted_residual_setting)
            )

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(
            last_channel * max(1.0, width_mult), round_nearest
        )
        features = [ConvBNReLU(3, input_channel, stride=2, norm_layer=norm_layer)]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(
                    block(
                        input_channel,
                        output_channel,
                        stride,
                        expand_ratio=t,
                        norm_layer=norm_layer,
                    )
                )
                input_channel = output_channel
        # building last several layers
        features.append(
            ConvBNReLU(
                input_channel, self.last_channel, kernel_size=1, norm_layer=norm_layer
            )
        )
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        self.pool = Lift(nn.AdaptiveAvgPool2d(1))

        # building classifier
        self.classifier = nn.Sequential(
            Lift(nn.Dropout(0.2)),
            Lift(nn.Linear(self.last_channel, num_classes)),
        )

        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                assert m.bias is None
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def _forward_impl(self, x):
        # This exists since TorchScript doesn't support inheritance, so the superclass method
        # (this one) needs to have a name other than `forward` that can be accessed in a subclass
        x, _ = self.features(x)
        # Cannot use "squeeze" as batch-size can be 1 => must use reshape with x.shape[0]
        # x = nn.functional.adaptive_avg_pool2d(x, 1).reshape(x.shape[1], -1)
        x = self.pool(x).reshape(x.shape[0], x.shape[1], -1)
        x = self.classifier(x)
        return x

    def forward(self, x):
        return self._forward_impl(x)


def mobilenet_v2(pretrained=False, progress=True, **kwargs):
    """
    Constructs a MobileNetV2 architecture from
    `"MobileNetV2: Inverted Residuals and Linear Bottlenecks" <https://arxiv.org/abs/1801.04381>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    model = MobileNetV2(**kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(
            model_urls["mobilenet_v2"], progress=progress
        )

        snn_state_dict = {}

        # compared to the ANN parameters, the modules are lifted
        # we modify the state dict accordingly here
        for key in state_dict:
            l = key.split(".")
            l.insert(-1, "lifted_module")
            new_key = ".".join(l)
            snn_state_dict[new_key] = state_dict[key]

        model.load_state_dict(snn_state_dict)

    return model



from typing import Tuple

import torch

from norse.torch.functional.heaviside import heaviside


class STDPState:
    """State of spike-timing-dependent plasticity (STDP).
    Parameters:
        t_pre (torch.Tensor): presynaptic spike trace
        t_post (torch.Tensor): postsynaptic spike trace
    """

    def __init__(self, t_pre: torch.Tensor, t_post: torch.Tensor):
        self.t_pre = t_pre
        self.t_post = t_post

    def decay(
        self,
        z_pre: torch.Tensor,
        z_post: torch.Tensor,
        tau_pre_inv: torch.Tensor,
        tau_post_inv: torch.Tensor,
        a_pre: torch.Tensor,
        a_post: torch.Tensor,
        dt: float = 0.001,
    ):
        """Decay function for STDP traces.
        Parameters:
            z_pre (torch.Tensor): presynaptic spikes
            z_post (torch.Tensor): postsynaptic spikes
            tau_pre_inv (torch.Tensor): inverse time-constant for the presynaptic trace
            tau_post (torch.Tensor): inverse time-constant for the postsynaptic trace
            a_pre (torch.Tensor): presynaptic trace
            a_post (torch.Tensor): postsynaptic trace
            dt (float): time-resolution
        """
        self.t_pre = self.t_pre + (dt * tau_pre_inv * (-self.t_pre + a_pre * z_pre))
        self.t_post = self.t_post + (
            dt * tau_post_inv * (-self.t_post + a_post * z_post)
        )


class STDPParameters:
    """STDP parameters.
    Parameters:
        a_pre (torch.Tensor): Contribution of presynaptic spikes to trace
        a_post (torch.Tensor): Contribution of postsynaptic spikes to trace
        tau_pre_inv (torch.Tensor): Inverse decay time constant of presynaptic spike trace in 1/s
        tau_post_inv (torch.Tensor): Inverse decay time constant of postsynaptic spike trace in 1/s
        w_min (torch.Tensor): Lower bound on synaptic weights (should be < w_max)
        w_max (torch.Tensor): Upper bound on synaptic weight (should be > w_min)
        eta_plus (torch.Tensor): Learning rate for synaptic potentiation (0 < eta_plus << 1)
        eta_minus (torch.Tensor): Learning rate for synaptic depression (0 < eta_minus << 1)
        stdp_algorithm (string): Algorithm for STDP updates. Options in {"additive","additive_step","multiplicative_pow","multiplicative_relu"}
        mu (torch.Tensor): Exponent for multiplicative STDP (0 <= mu <= 1)
        hardbound (boolean): Impose hardbounds by clipping (recommended unles eta_* << 1)
        convolutional (boolean): Convolutional weighting kernel
        stride (int): Stride for convolution
        padding (int): Padding for convolution
        dilation (int): Dilation for convolution
    """

    def Additive(self):
        return (
            torch.tensor(0.0),
            lambda w: self.eta_plus,
            lambda w: self.eta_minus,
        )

    def __init__(
        self,
        a_pre: torch.Tensor = torch.as_tensor(1.0),
        a_post: torch.Tensor = torch.as_tensor(1.0),
        tau_pre_inv: torch.Tensor = torch.as_tensor(1.0 / 50e-3),
        tau_post_inv: torch.Tensor = torch.as_tensor(1.0 / 50e-3),
        w_min: torch.Tensor = 0.0,
        w_max: torch.Tensor = 1.0,
        eta_plus: torch.Tensor = 1e-3,
        eta_minus: torch.Tensor = 1e-3,
        stdp_algorithm: str = "additive",
        mu: torch.Tensor = 0.0,
        hardbound: bool = True,
        convolutional: bool = False,
        stride: int = 1,
        padding: int = 0,
        dilation: int = 1,
    ):
        self.a_pre = a_pre
        self.a_post = a_post
        self.tau_pre_inv = tau_pre_inv
        self.tau_post_inv = tau_post_inv
        self.w_min = w_min
        self.w_max = w_max
        self.eta_plus = eta_plus
        self.eta_minus = eta_minus

        self.stdp_algorithm = stdp_algorithm
        if self.stdp_algorithm == "additive":
            self.mu = torch.tensor(0.0)
            self.A_plus = lambda w: self.eta_plus
            self.A_minus = lambda w: self.eta_minus
        elif self.stdp_algorithm == "additive_step":
            self.mu = torch.tensor(0.0)
            self.A_plus = lambda w: self.eta_plus * heaviside(self.w_max - w)
            self.A_minus = lambda w: self.eta_minus * heaviside(w - self.w_min)
        elif self.stdp_algorithm == "multiplicative_pow":
            self.mu = torch.tensor(mu)
            self.A_plus = lambda w: self.eta_plus * torch.pow(self.w_max - w, self.mu)
            self.A_minus = lambda w: self.eta_minus * torch.pow(w - self.w_min, self.mu)
        elif self.stdp_algorithm == "multiplicative_relu":
            self.mu = torch.tensor(1.0)
            self.A_plus = lambda w: self.eta_plus * torch.nn.functional.relu(
                self.w_max - w
            )
            self.A_minus = lambda w: self.eta_minus * torch.nn.functional.relu(
                w - self.w_min
            )

        # Hard bounds
        self.hardbound = hardbound
        if self.hardbound:
            self.bounding_func = lambda w: torch.clamp(w, w_min, w_max)
        else:
            self.bounding_func = lambda w: w

        # Conv2D
        self.convolutional = convolutional
        if self.convolutional:
            self.stride = stride
            self.padding = padding
            self.dilation = dilation


def stdp_step_linear(
    z_pre: torch.Tensor,
    z_post: torch.Tensor,
    w: torch.Tensor,
    state_stdp: STDPState = STDPState,
    p_stdp: STDPParameters = STDPParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, STDPState]:
    """STDP step for a FF LIF layer.
    Input:
        z_pre (torch.Tensor): Presynaptic activity z: {0,1} -> {no spike, spike}
        z_post (torch.Tensor): Postsynaptic activity z: {0,1} -> {no spike, spike}
        w (torch.Tensor): Weight tensor connecting the pre- and postsynaptic layers
        state_stdp (STDPState): STDP state
        p_stdp (STDPParameters): Parameters of STDP
        dt (float): Time-resolution
    Output:
        w (torch.tensor): Updated synaptic weights
        state_stdp (STDPState): Updated STDP state
    """

    # Update STDP traces
    state_stdp.decay(
        z_pre,
        z_post,
        p_stdp.tau_pre_inv,
        p_stdp.tau_post_inv,
        p_stdp.a_pre,
        p_stdp.a_post,
        dt,
    )

    # STDP weight update
    dw_plus = p_stdp.A_plus(w) * torch.einsum("bi,bj->ij", z_post, state_stdp.t_pre)
    dw_minus = p_stdp.A_minus(w) * torch.einsum("bi,bj->ij", state_stdp.t_post, z_pre)

    w = w + (dw_plus - dw_minus)

    # Bound checking
    w = p_stdp.bounding_func(w)

    return (w, state_stdp)


def stdp_step_conv2d(
    z_pre: torch.Tensor,
    z_post: torch.Tensor,
    w: torch.Tensor,
    state_stdp: STDPState = STDPState,
    p_stdp: STDPParameters = STDPParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, STDPState]:
    """STDP step for a conv2d LIF layer.
    Input:
        z_pre (torch.tensor): Presynaptic activity z: {0,1} -> {no spike, spike}
        z_post (torch.tensor): Postsynaptic activity z: {0,1} -> {no spike, spike}
        w (torch.Tensor): Weight tensor connecting the pre- and postsynaptic layers
        state_stdp (STDPState): STDP state
        p_stdp (STDPParameters): Parameters of STDP
        dt (float): Time-resolution
    Output:
        w (torch.tensor): Updated synaptic weights
        state_stdp (STDPState): Updated STDP state
    """

    # Update STDP traces
    state_stdp.decay(
        z_pre,
        z_post,
        p_stdp.tau_pre_inv,
        p_stdp.tau_post_inv,
        p_stdp.a_pre,
        p_stdp.a_post,
        dt,
    )

    # Unfolding the convolution
    # Get shape information
    batch_size = z_pre.shape[0]
    out_channels, _, kernel_h, kernel_w = w.shape

    # Unfold presynaptic trace
    state_stdp_t_pre_uf = torch.nn.functional.unfold(
        state_stdp.t_pre,
        (kernel_h, kernel_w),
        dilation=p_stdp.dilation,
        padding=p_stdp.padding,
        stride=p_stdp.stride,
    )
    # Unfold presynaptic receptive fields
    z_pre_uf = torch.nn.functional.unfold(
        z_pre,
        (kernel_h, kernel_w),
        dilation=p_stdp.dilation,
        padding=p_stdp.padding,
        stride=p_stdp.stride,
    )

    # STDP weight update
    dw_plus = (
        p_stdp.A_plus(w)
        * torch.einsum(
            "bik,bjk->ij",
            z_post.view(batch_size, out_channels, -1),
            state_stdp_t_pre_uf,
        ).view(w.shape)
    )

    dw_minus = (
        p_stdp.A_minus(w)
        * torch.einsum(
            "bik,bjk->ij",
            state_stdp.t_post.view(batch_size, out_channels, -1),
            z_pre_uf,
        ).view(w.shape)
    )

    w = w + (dw_plus - dw_minus)

    # Bound checking
    w = p_stdp.bounding_func(w)

    return (w, state_stdp)



import torch
from torch import nn
import norse.torch as norse
from ..lsnn import LSNNCell, LSNNLayer
from ..lif import LIFCell, LIFLayer
from ..lift import Lift
from ..sequential import SequentialState


def test_state_sequence():
    d = torch.ones(10, 1, 20)
    l = LIFLayer(20, 6)
    z, s = SequentialState(l)(d)
    assert z.shape == (10, 1, 6)
    assert s[0].v.shape == (1, 6)


def test_state_sequence_list():
    d = torch.ones(10, 1, 20)
    l1 = LIFLayer(20, 6)
    l2 = LIFLayer(6, 2)
    s = [None, None]
    z, s = SequentialState(l1, l2)(d, s)
    assert z.shape == (10, 1, 2)
    assert s[1].v.shape == (1, 2)


def test_state_sequence_norse():
    d = torch.ones(10, 2, 10)
    l1 = LIFLayer(10, 5)
    l2 = LSNNLayer(LIFCell, 5, 1)
    z, (s1, s2) = SequentialState(l1, l2)(d)
    assert z.shape == (10, 2, 1)
    assert s1.v.shape == (2, 5)
    assert s2.v.shape == (2, 1)


def test_state_sequence_mix():
    d = torch.ones(10, 3, 20)
    l1 = LIFLayer(20, 10)
    l2 = torch.nn.RNN(10, 4, 2)  # 2 layers
    l3 = LSNNLayer(LSNNCell, 4, 1)
    state = [None, torch.randn(2, 3, 4), None]
    z, (s1, s2, s3) = SequentialState(l1, l2, l3)(d, state)
    assert z.shape == (10, 3, 1)
    assert s1.v.shape == (3, 10)
    assert s2.shape == (2, 3, 4)
    assert s3.v.shape == (3, 1)


def test_state_sequence_conv():
    data = torch.ones(1, 8, 16, 4, 4)  # (timestep, minibatch, channels, x, y)
    model = SequentialState(
        Lift(torch.nn.Conv2d(16, 8, 3)),  # (1, 8, 8, 2, 2)
        torch.nn.Flatten(2),  # (1, 8, 32)
        LSNNLayer(LSNNCell, 32, 6),  # (1, 8, 6)
        torch.nn.RNN(6, 4, 2),  # (1, 6, 4) with 2 recurrent layers
        LIFLayer(4, 1),  # (1, 4, 1)
    )
    model(data)


def test_backprop_works():
    model = SequentialState(
        norse.LSNNLayer(norse.LSNNCell, 1, 2),
        norse.LSNNLayer(norse.LSNNCell, 2, 3),
        norse.LSNNLayer(norse.LSNNCell, 3, 3),
        norse.LSNNLayer(norse.LSNNCell, 3, 2),
        nn.Flatten(),
        norse.Lift(torch.nn.Linear(4, 2)),
        norse.Lift(torch.nn.Linear(2, 1)),
    )
    optimizer = torch.optim.Adam(model.parameters(), lr=0.0001)
    loss_func = nn.MSELoss()
    data = torch.ones(3, 1, 2, 1)
    target = torch.ones(1, 2)
    state = None
    for x in data:
        out, state = model(x, state)
        loss = loss_func(out, target)
        optimizer.zero_grad()  # clear gradients for this training step
        loss.backward()  # backpropagation, compute gradients
        optimizer.step()
        state = [s.detach() if s is not None else None for s in state]



import torch

from norse.torch.functional.superspike import super_fn


def logical_and(x, y):
    """Computes a logical and provided x and y are bitvectors."""
    return x * y


def logical_xor(x, y):
    """Computes a logical xor provided x and y are bitvectors."""
    return torch.abs(x - y)


def logical_or(x, y):
    """Computes a logical or provided x and y are bitvectors."""
    return super_fn(x + y)


def muller_c(y_prev, x_1, x_2):
    """Computes the muller-c element next state provided x_1 and x_2 are bitvectors
    and y_prev is the previous state."""
    return super_fn(x_1 * x_2 + (x_1 + x_2) * y_prev)


def posedge_detector(z, z_prev):
    """Determines whether a transition from 0 to 1 has occured
    providing that z and z_prev are bitvectors
    """
    return (1 - z_prev) * z



import torch
import torch.jit
import numpy as np
from typing import Optional, Tuple

from ..functional.leaky_integrator import (
    li_step,
    li_feed_forward_step,
    LIState,
    LIParameters,
)


class LICell(torch.nn.Module):
    r"""Cell for a leaky-integrator.
    More specifically it implements a discretized version of the ODE

    .. math::

        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}


    and transition equations

    .. math::
        i = i + w i_{\text{in}}

    Parameters:
        input_size (int): Size of the input. Also known as the number of input features.
        hidden_size (int): Size of the hidden state. Also known as the number of input features.
        p (LIParameters): parameters of the leaky integrator
        dt (float): integration timestep to use
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        p: LIParameters = LIParameters(),
        dt: float = 0.001,
    ):
        super(LICell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(output_size, input_size) / np.sqrt(input_size)
        )
        self.p = p
        self.dt = dt
        self.output_size = output_size

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIState] = None
    ) -> Tuple[torch.Tensor, LIState]:
        if state is None:
            state = LIState(
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    (input_tensor.shape[0], self.output_size),
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return li_step(
            input_tensor,
            state,
            self.input_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFeedForwardCell(torch.nn.Module):
    r"""Cell for a leaky-integrator.
    More specifically it implements a discretized version of the ODE

    .. math::

        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}


    and transition equations

    .. math::
        i = i + w i_{\text{in}}

    Parameters:
        p (LIParameters): parameters of the leaky integrator
        dt (float): integration timestep to use
    """

    def __init__(self, p: LIParameters = LIParameters(), dt: float = 0.001):
        super(LIFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIState] = None
    ) -> Tuple[torch.Tensor, LIState]:
        if state is None:
            state = LIState(
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    *input_tensor.shape,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return li_feed_forward_step(input_tensor, state, p=self.p, dt=self.dt)



import torch

from norse.torch.functional.leaky_integrator import li_step, LIState
from norse.torch.functional.lif import (
    lif_step,
    lif_feed_forward_step,
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
)
from norse.torch.functional.lift import lift


def test_lift_without_state_or_parameters():
    data = torch.ones(3, 2, 1)
    lifted = lift(lif_feed_forward_step)
    z, s = lifted(data)
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)
    assert s.i.shape == (2, 1)


def test_lift_with_state_without_parameters():
    data = torch.ones(3, 2, 1)
    lifted = lift(lif_feed_forward_step)
    z, s = lifted(
        data,
        state=LIFFeedForwardState(torch.zeros_like(data[0]), torch.zeros_like(data[0])),
    )
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)
    assert s.i.shape == (2, 1)


def test_lift_without_state_with_parameters():
    data = torch.ones(3, 2, 1)
    lifted = lift(
        lif_feed_forward_step, p=LIFParameters(v_th=torch.as_tensor(0.3), method="tanh")
    )
    z, s = lifted(data)
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)
    assert s.i.shape == (2, 1)


def test_lift_with_state_and_parameters():
    data = torch.ones(3, 2, 1)
    lifted = lift(
        lif_feed_forward_step, p=LIFParameters(v_th=torch.as_tensor(0.3), method="tanh")
    )
    z, s = lifted(
        data,
        state=LIFFeedForwardState(torch.zeros_like(data[0]), torch.zeros_like(data[0])),
    )
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)
    assert s.i.shape == (2, 1)


def test_lift_with_lift_step():
    data = torch.ones(3, 2, 1)
    lifted = lift(lif_step)
    z, s = lifted(
        data,
        state=LIFState(
            v=torch.zeros(2, 1),
            i=torch.zeros(2, 1),
            z=torch.zeros(2, 1),
        ),
        input_weights=torch.ones(1, 1),
        recurrent_weights=torch.ones(1, 1),
    )
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)


def test_lift_with_leaky_integrator():
    data = torch.ones(3, 2, 1)
    lifted = lift(li_step)
    z, s = lifted(
        data,
        state=LIState(
            v=torch.zeros(2, 1),
            i=torch.zeros(2, 1),
        ),
        input_weights=torch.ones(1, 1),
    )
    assert z.shape == (3, 2, 1)
    assert s.v.shape == (2, 1)



from typing import NamedTuple, Tuple

import torch

from norse.torch.functional.threshold import threshold


class CobaLIFState(NamedTuple):
    """State of a conductance based LIF neuron.

    Parameters:
        z (torch.Tensor): recurrent spikes
        v (torch.Tensor): membrane potential
        g_e (torch.Tensor): excitatory input conductance
        g_i (torch.Tensor): inhibitory input conductance
    """

    z: torch.Tensor
    v: torch.Tensor
    g_e: torch.Tensor
    g_i: torch.Tensor


default_bio_state = CobaLIFState(z=0.0, v=-65.0, g_e=0.0, g_i=0.0)


class CobaLIFParameters(NamedTuple):
    """Parameters of conductance based LIF neuron.

    Parameters:
        tau_syn_exc_inv (torch.Tensor): inverse excitatory synaptic input
                                        time constant
        tau_syn_inh_inv (torch.Tensor): inverse inhibitory synaptic input
                                        time constant
        c_m_inv (torch.Tensor): inverse membrane capacitance
        g_l (torch.Tensor): leak conductance
        e_rev_I (torch.Tensor): inhibitory reversal potential
        e_rev_E (torch.Tensor): excitatory reversal potential
        v_rest (torch.Tensor): rest membrane potential
        v_reset (torch.Tensor): reset membrane potential
        v_thresh (torch.Tensor): threshold membrane potential
        method (str): method to determine the spike threshold
                      (relevant for surrogate gradients)
        alpha (float): hyper parameter to use in surrogate gradient computation
    """

    tau_syn_exc_inv: torch.Tensor = torch.as_tensor(1.0 / 5)
    tau_syn_inh_inv: torch.Tensor = torch.as_tensor(1.0 / 5)
    c_m_inv: torch.Tensor = torch.as_tensor(1 / 0.2)
    g_l: torch.Tensor = torch.as_tensor(1 / 20 * 1 / 0.2)
    e_rev_I: torch.Tensor = torch.as_tensor(-100)
    e_rev_E: torch.Tensor = torch.as_tensor(60)
    v_rest: torch.Tensor = torch.as_tensor(-20)
    v_reset: torch.Tensor = torch.as_tensor(-70)
    v_thresh: torch.Tensor = torch.as_tensor(-10)
    method: str = "super"
    alpha: float = 100.0


default_bio_parameters = CobaLIFParameters(
    tau_syn_exc_inv=1 / 0.3,
    tau_syn_inh_inv=1 / 0.5,
    e_rev_E=0.0,
    e_rev_I=-70.0,
    v_thresh=-50.0,
    v_reset=-65.0,
    v_rest=-65.0,
)


def coba_lif_step(
    input_tensor: torch.Tensor,
    state: CobaLIFState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: CobaLIFParameters = CobaLIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, CobaLIFState]:
    """Euler integration step for a conductance based LIF neuron.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (CobaLIFState): current state of the neuron
        input_weights (torch.Tensor): input weights
            (sign determines  contribution to inhibitory / excitatory input)
        recurrent_weights (torch.Tensor): recurrent weights
            (sign determines contribution to inhibitory / excitatory input)
        p (CobaLIFParameters): parameters of the neuron
        dt (float): Integration time step
    """
    dg_e = -dt * p.tau_syn_exc_inv * state.g_e
    g_e = state.g_e + dg_e
    dg_i = -dt * p.tau_syn_inh_inv * state.g_i
    g_i = state.g_i + dg_i

    g_e = g_e + torch.nn.functional.linear(
        input_tensor, torch.nn.functional.relu(input_weights)
    )
    g_i = g_i + torch.nn.functional.linear(
        input_tensor, torch.nn.functional.relu(-input_weights)
    )

    g_e = g_e + torch.nn.functional.linear(
        state.z, torch.nn.functional.relu(recurrent_weights)
    )
    g_i = g_i + torch.nn.functional.linear(
        state.z, torch.nn.functional.relu(-recurrent_weights)
    )

    dv = (
        dt
        * p.c_m_inv
        * (
            p.g_l * (p.v_rest - state.v)
            + g_e * (p.e_rev_E - state.v)
            + g_i * (p.e_rev_I - state.v)
        )
    )
    v = state.v + dv

    z_new = threshold(v - p.v_thresh, p.method, p.alpha)
    v = (1 - z_new) * v + z_new * p.v_reset
    return z_new, CobaLIFState(z_new, v, g_e, g_i)


class CobaLIFFeedForwardState(NamedTuple):
    """State of a conductance based feed forward LIF neuron.

    Parameters:
        v (torch.Tensor): membrane potential
        g_e (torch.Tensor): excitatory input conductance
        g_i (torch.Tensor): inhibitory input conductance
    """

    v: torch.Tensor
    g_e: torch.Tensor
    g_i: torch.Tensor


def coba_lif_feed_forward_step(
    input_tensor: torch.Tensor,
    state: CobaLIFFeedForwardState,
    p: CobaLIFParameters = CobaLIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, CobaLIFFeedForwardState]:
    """Euler integration step for a conductance based LIF neuron.

    Parameters:
        input_tensor (torch.Tensor): synaptic input
        state (CobaLIFFeedForwardState): current state of the neuron
        p (CobaLIFParameters): parameters of the neuron
        dt (float): Integration time step
    """
    dg_e = -dt * p.tau_syn_exc_inv * state.g_e
    g_e = state.g_e + dg_e
    dg_i = -dt * p.tau_syn_inh_inv * state.g_i
    g_i = state.g_i + dg_i

    g_e = g_e + torch.nn.functional.relu(input_tensor)
    g_i = g_i + torch.nn.functional.relu(-input_tensor)

    dv = (
        dt
        * p.c_m_inv
        * (
            p.g_l * (p.v_rest - state.v)
            + g_e * (p.e_rev_E - state.v)
            + g_i * (p.e_rev_I - state.v)
        )
    )
    v = state.v + dv

    z_new = threshold(v - p.v_thresh, p.method, p.alpha)
    v = (1 - z_new) * v + z_new * p.v_reset
    return z_new, CobaLIFFeedForwardState(v, g_e, g_i)



from typing import Tuple

import torch

from norse.torch.functional.lif import (
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
    lif_step,
    lif_feed_forward_step,
)


def lif_mc_step(
    input_tensor: torch.Tensor,
    state: LIFState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    g_coupling: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFState]:
    """Computes a single euler-integration step of a LIF multi-compartment
    neuron-model.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LIFState): current state of the neuron
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """
    v_new = state.v + dt * torch.nn.functional.linear(state.v, g_coupling)
    return lif_step(
        input_tensor,
        LIFState(state.z, v_new, state.i),
        input_weights,
        recurrent_weights,
        p,
        dt,
    )


def lif_mc_feed_forward_step(
    input_tensor: torch.Tensor,
    state: LIFFeedForwardState,
    g_coupling: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFFeedForwardState]:
    """Computes a single euler-integration feed forward step of a LIF
    multi-compartment neuron-model.

    Parameters:
        input_tensor (torch.Tensor): the (weighted) input spikes at the
                              current time step
        s (LIFFeedForwardState): current state of the neuron
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """
    v_new = state.v + dt * torch.nn.functional.linear(state.v, g_coupling)
    return lif_feed_forward_step(
        input_tensor, LIFFeedForwardState(v_new, state.i), p, dt
    )



from absl import app, flags
import datetime
import logging
import numpy as np
import torch
import torch.utils.data

from norse.torch.functional.encode import poisson_encode
from norse.torch.module.leaky_integrator import LICell
from norse.torch.module.lsnn import LSNNCell, LSNNParameters
from norse.torch.module.lif import LIFCell, LIFParameters

flags.DEFINE_integer("batch_size", 64, "Batch size.")
flags.DEFINE_enum("device", "cpu", ["cpu", "cuda"], "Device to use by pytorch.")
flags.DEFINE_integer("device_number", 0, "Index of the CUDA device to use, if at all.")
flags.DEFINE_float("dt", 0.001, "Time change per simulation step.")
flags.DEFINE_integer("epochs", 20, "Number of epochs.")
flags.DEFINE_float("learning_rate", 2e-3, "Learning rate to use.")
flags.DEFINE_integer(
    "log_interval", 10, "In which intervals to display learning progress."
)
flags.DEFINE_enum(
    "model",
    "super",
    ["super", "tanh", "circ", "logistic", "circ_dist"],
    "Model to use for training.",
)
flags.DEFINE_enum(
    "neuron_model",
    "lsnn",
    ["lsnn", "lif"],
    "Neuron model to use in network.",
)
flags.DEFINE_enum(
    "optimizer", "adam", ["adam", "sgd"], "Optimizer to use for training."
)
flags.DEFINE_boolean("plot", False, "Do intermediate plots during test sets.")
flags.DEFINE_enum(
    "plot_output",
    "show",
    ["show", "file"],
    "How to output plots: display on screen (show) or save to file (file).",
)

flags.DEFINE_integer(
    "poisson_rate",
    50,
    "Number of spikes per second, drawn from a poisson distribution.",
)
flags.DEFINE_integer(
    "population_size", 5, "Number of neurons per input bit (population encoded)."
)
flags.DEFINE_integer(
    "random_seed", int(torch.randint(high=100000, size=(1,))[0]), "Random seed."
)
flags.DEFINE_integer("samples", 1000, "Number of samples to use.")
flags.DEFINE_boolean("save_model", True, "Save the model after training every epoch.")
flags.DEFINE_integer("seq_length", 200, "Number of time steps per experiment step.")
flags.DEFINE_integer(
    "seq_steps", 24, "Number of steps in each experiment (should be at least 8)."
)
flags.DEFINE_integer(
    "seq_repetitions",
    24,
    "Number of times a sequence is repeated.",
)
flags.DEFINE_float("weight_decay", 1e-5, "Weight decay (L2 regularisation penalty).")


class MemoryNet(torch.nn.Module):
    def __init__(
        self,
        input_features,
        output_features,
        seq_length,
        is_lsnn,
        dt=0.01,
        model="super",
    ):
        super(MemoryNet, self).__init__()
        self.input_features = input_features
        self.output_features = output_features
        self.seq_length = seq_length
        self.is_lsnn = is_lsnn
        if is_lsnn:
            p = LSNNParameters(method=model)
            self.layer = LSNNCell(input_features, input_features, p, dt=dt)
        else:
            p = LIFParameters(method=model)
            self.layer = LIFCell(input_features, input_features, dt=dt)
        self.dropout = torch.nn.Dropout(p=0.2)
        self.readout = LICell(input_features, output_features)

    def forward(self, x):
        batch_size = x.shape[0]

        sl = None
        sr = None
        seq_spikes = []
        step_spikes = []
        seq_readouts = []
        step_readouts = []
        for index, x_step in enumerate(x.unbind(1)):
            spikes, sl = self.layer(x_step, sl)
            seq_spikes.append(spikes)
            spikes = self.dropout(spikes)
            _, sr = self.readout(spikes, sr)
            seq_readouts.append(sr.v)
            if (index + 1) % self.seq_length == 0:
                step_spikes.append(torch.stack(seq_spikes))
                seq_spikes = []
                step_readouts.append(torch.stack(seq_readouts))
                seq_readouts = []
        spikes = torch.cat(step_spikes)
        readouts = torch.stack(step_readouts)
        return readouts, spikes


class MemoryDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        samples,
        steps,
        seq_length,
        seq_repetitions,
        population_size,
        device,
        poisson_rate=1,
        dt=0.001,
        generator=torch.default_generator,
    ):
        self.samples = samples
        self.steps = steps
        self.seq_length = seq_length
        self.seq_repetitions = seq_repetitions
        self.population_size = population_size
        self.poisson_rate = poisson_rate
        self.device = device
        self.dt = dt
        self.generator = generator

        self.store_indices = torch.randint(
            low=0,
            high=steps // 2,
            size=(samples, seq_repetitions),
            generator=generator,
        )
        self.recall_indices = torch.randint(
            low=steps // 2,
            high=steps,
            size=(samples, seq_repetitions),
            generator=generator,
        )

    def __len__(self):
        return self.samples

    def _generate_sequence(self, idx, rep_idx):
        data_pattern = torch.stack(
            [torch.randperm(2, generator=self.generator) for _ in range(self.steps)]
        ).byte()
        store_index = self.store_indices[idx][rep_idx]
        recall_index = self.recall_indices[idx][rep_idx]
        store_pattern = torch.zeros((self.steps, 1)).byte()
        recall_pattern = store_pattern.clone()
        label_pattern = torch.zeros((self.steps, 2)).byte()

        store_pattern[store_index] = 1
        recall_pattern[recall_index] = 1
        label_class = data_pattern[store_index].byte()
        label_pattern[store_index] = label_class
        label_pattern[recall_index] = label_class
        data_pattern[recall_index] = torch.zeros(2)

        input_pattern = torch.cat((data_pattern, store_pattern, recall_pattern), dim=1)
        input_pattern = input_pattern.repeat_interleave(self.population_size, dim=1)
        encoded = poisson_encode(
            input_pattern,
            seq_length=self.seq_length,
            f_max=self.poisson_rate,
            dt=self.dt,
        )
        encoded = torch.cat(encoded.chunk(self.steps, dim=1)).squeeze()
        return encoded.to(self.device), label_pattern.to(self.device)

    def __getitem__(self, idx):
        repetitions = [
            self._generate_sequence(idx, i) for i in range(self.seq_repetitions)
        ]
        return (
            torch.cat([rep[0] for rep in repetitions]),
            torch.cat([rep[1] for rep in repetitions]),
        )


def _memory_accuracy_loss(xs, ys):
    xs = torch.stack(xs.mean(1).softmax(2).unbind(dim=1), dim=0)

    loss = torch.nn.functional.binary_cross_entropy(xs, ys.float())

    target = ys.argmax(1).min(1).indices
    batch_indices = ys.argmax(1).min(1).values
    recall_values = torch.stack(
        [xs[idx][batch_index] for idx, batch_index in enumerate(batch_indices)]
    ).argmax(1)
    return (target == recall_values).float().sum(), loss


def _plot_run(
    xs, ys, readouts, spikes, epoch, total_epochs, accuracy, file_prefix=None
):
    """
    Only plots the first batch event
    """

    def spikes_to_events(spike_list):
        return [[x for x, spike in enumerate(ts) if spike > 0] for ts in spike_list.T]

    try:
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec

        # Setup grid
        plt.title(f"Accuracy: {accuracy} (epoch {epoch}/{total_epochs}")
        gridspec.GridSpec(4, 1)
        plt.subplot2grid((4, 1), (0, 0), rowspan=3)
        input_events = spikes_to_events(xs[0])
        network_events = spikes_to_events(spikes[:, 0])
        plt.eventplot(
            network_events + input_events[::-1],
            linewidth=1,
            linelengths=0.8,
            color="black",
        )
        yticks = [x for x in range(len(network_events))] + [22.5, 27.5, 32.5, 37.5]
        y_labels = [""] * len(network_events)
        y_labels.append("Recall")
        y_labels.append("Store")
        y_labels.append("1")
        y_labels.append("0")
        plt.gca().set_yticks(yticks)
        plt.gca().set_yticklabels(y_labels)

        readout_events = readouts[:, :, 0].detach().reshape(-1, 2).softmax(1)
        readout_sum = (readout_events[:, 0] - readout_events[:, 1]).cpu().numpy()

        plt.subplot2grid((4, 1), (3, 0), rowspan=1)
        plt.gca().set_ylim(-1.1, 1.1)
        plt.gca().set_yticks([-1, 0, 1])
        plt.plot(readout_sum, color="black")

        plt.tight_layout()
        if file_prefix:
            plt.savefig(f"{file_prefix}_{epoch}.png")
        else:
            plt.show()
        plt.close()
    except ImportError as e:
        logging.warning("Plotting failed: Cannot import matplotlib: " + str(e))


def train(
    model,
    data_loader,
    optimizer,
    epoch,
    total_epochs,
    log_interval=1e10,
    writer=None,
):
    model.train()
    losses = []
    optimizer.zero_grad()
    step = 0

    for batch_idx, (xs, ys) in enumerate(data_loader):
        optimizer.zero_grad()
        readouts, _ = model(xs)
        accuracy, loss = _memory_accuracy_loss(readouts, ys)
        loss.backward()

        optimizer.step()
        step += 1

        logging.info(
            "Train Epoch: {}/{} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                epoch,
                total_epochs,
                batch_idx * len(xs),
                len(data_loader.dataset),
                100.0 * batch_idx / len(data_loader),
                loss.item(),
            )
        )

        if step % log_interval == 0 and writer:
            try:
                writer.add_scalar("Loss/train", loss.item(), step)
                writer.add_scalar("Accuracy/train", accuracy.item(), step)

                for tag, value in model.named_parameters():
                    tag = tag.replace(".", "/")
                    writer.add_histogram(tag, value.data.cpu().numpy(), step)
                    writer.add_histogram(
                        tag + "/grad", value.grad.data.cpu().numpy(), step
                    )
            except ValueError as e:
                logging.warning("Error when writing to tensorboard: " + str(e))

        losses.append(loss.item())

    mean_loss = torch.mean(torch.tensor(losses).float())
    return losses, mean_loss


def test(
    model,
    method,
    test_loader,
    epoch,
    total_epochs,
    plot=False,
    plot_file_prefix=None,
    writer=None,
):
    model.eval()
    test_loss = 0
    correct = 0
    did_plot = False
    runs = []

    with torch.no_grad():
        for xs, ys in test_loader:
            readouts, spikes = model(xs)
            if plot:
                runs.append((xs, ys, readouts, spikes))
            accuracy, loss = _memory_accuracy_loss(readouts, ys)
            test_loss += loss.item()
            correct += accuracy

            if plot and not did_plot:
                _plot_run(
                    xs,
                    ys,
                    readouts,
                    spikes,
                    epoch,
                    total_epochs,
                    accuracy,
                    plot_file_prefix,
                )
                did_plot = True

    test_loss /= len(test_loader.dataset)

    accuracy = 100.0 * correct / len(test_loader.dataset)
    logging.info(
        f"\nTest set {method}: Average loss: {test_loss:.4f}, \
            Accuracy: {correct}/{len(test_loader.dataset)} ({accuracy:.0f}%)\n"
    )
    if writer:
        writer.add_scalar("Loss/test", test_loss, epoch)
        writer.add_scalar("Accuracy/test", accuracy, epoch)

    return test_loss, accuracy


def save(path, epoch, model, optimizer, is_best=False):
    torch.save(
        {
            "epoch": epoch + 1,
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "is_best": is_best,
        },
        path,
    )


def main(args):
    try:
        from torch.utils.tensorboard import SummaryWriter

        writer = SummaryWriter()
    except ImportError:
        logging.info(
            "Disabling logging due to missing tensorboard dependency. Install tensorboard to enable logging."
        )
        writer = None

    FLAGS = flags.FLAGS

    input_features = (
        4 * FLAGS.population_size
    )  # (two bits + store + recall) * population size
    output_features = 2
    batch_size = FLAGS.batch_size
    epochs = FLAGS.epochs

    generator = torch.random.manual_seed(FLAGS.random_seed)

    if FLAGS.device == "cuda":
        # Workaround for https://github.com/pytorch/pytorch/issues/21819
        torch.cuda.set_device(FLAGS.device_number)

    model = MemoryNet(
        input_features,
        output_features,
        seq_length=FLAGS.seq_length,
        is_lsnn=FLAGS.neuron_model == "lsnn",
        dt=FLAGS.dt,
    ).to(FLAGS.device)

    if FLAGS.optimizer == "sgd":
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=FLAGS.learning_rate,
            momentum=0.9,
            weight_decay=FLAGS.weight_decay,
        )
    elif FLAGS.optimizer == "adam":
        optimizer = torch.optim.Adam(
            model.parameters(), lr=FLAGS.learning_rate, weight_decay=FLAGS.weight_decay
        )

    file_prefix = f"memory_{datetime.datetime.strftime(datetime.datetime.now(), '%Y%m%d_%Hh%M')}_{FLAGS.neuron_model}"

    dataset = MemoryDataset(
        FLAGS.samples,
        FLAGS.seq_steps,
        FLAGS.seq_length,
        FLAGS.seq_repetitions,
        FLAGS.population_size,
        device=FLAGS.device,
        poisson_rate=FLAGS.poisson_rate,
        dt=FLAGS.dt,
        generator=generator,
    )
    dataset_test = MemoryDataset(
        FLAGS.samples // 5,
        FLAGS.seq_steps,
        FLAGS.seq_length,
        FLAGS.seq_repetitions,
        FLAGS.population_size,
        device=FLAGS.device,
        poisson_rate=FLAGS.poisson_rate,
        dt=FLAGS.dt,
        generator=generator,
    )
    loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size)
    loader_test = torch.utils.data.DataLoader(dataset_test, batch_size=batch_size)

    training_losses = []
    mean_losses = []
    test_losses = []
    accuracies = []
    for epoch_index in range(epochs):
        training_loss, mean_loss = train(
            model,
            loader,
            optimizer,
            epoch_index,
            epochs,
            log_interval=FLAGS.log_interval,
            writer=writer,
        )
        test_loss, accuracy = test(
            model,
            FLAGS.model,
            loader_test,
            epoch_index,
            epochs,
            plot=FLAGS.plot,
            plot_file_prefix=(file_prefix if FLAGS.plot_output == "file" else None),
            writer=writer,
        )

        training_losses += training_loss
        mean_losses.append(mean_loss)
        test_losses.append(test_loss)
        accuracies.append(accuracy)

        max_accuracy = np.max(np.array(accuracies))

        if FLAGS.save_model:
            model_path = f"{file_prefix}_epoch_{epoch_index}.pt"
            save(
                model_path,
                model=model,
                optimizer=optimizer,
                epoch=epoch_index,
                is_best=accuracy > max_accuracy,
            )

    np.save("training_losses.npy", np.array(training_losses))
    np.save("mean_losses.npy", np.array(mean_losses))
    np.save("test_losses.npy", np.array(test_losses))
    np.save("accuracies.npy", np.array(accuracies))
    model_path = f"{file_prefix}_final.pt"
    save(
        model_path,
        epoch=epoch_index,
        model=model,
        optimizer=optimizer,
        is_best=accuracy > max_accuracy,
    )
    if writer:
        writer.close()


if __name__ == "__main__":
    app.run(main)



import torch

from norse.torch.functional.lif import LIFParameters
from norse.torch.module.leaky_integrator import LICell
from norse.torch.module.lif import LIFFeedForwardCell


class ConvNet(torch.nn.Module):
    def __init__(
        self, num_channels=1, feature_size=28, method="super", dtype=torch.float
    ):
        super(ConvNet, self).__init__()
        self.features = int(((feature_size - 4) / 2 - 4) / 2)
        self.conv1 = torch.nn.Conv2d(num_channels, 20, 5, 1)
        self.conv2 = torch.nn.Conv2d(20, 50, 5, 1)
        self.fc1 = torch.nn.Linear(self.features * self.features * 50, 500)
        self.out = LICell(500, 10)
        self.lif0 = LIFFeedForwardCell(
            p=LIFParameters(method=method, alpha=100.0),
        )
        self.lif1 = LIFFeedForwardCell(
            p=LIFParameters(method=method, alpha=100.0),
        )
        self.lif2 = LIFFeedForwardCell(p=LIFParameters(method=method, alpha=100.0))
        self.dtype = dtype

    def forward(self, x):
        seq_length = x.shape[0]
        batch_size = x.shape[1]

        # specify the initial states
        s0 = None
        s1 = None
        s2 = None
        so = None

        voltages = torch.zeros(
            seq_length, batch_size, 10, device=x.device, dtype=self.dtype
        )

        for ts in range(seq_length):
            z = self.conv1(x[ts, :])
            z, s0 = self.lif0(z, s0)
            z = torch.nn.functional.max_pool2d(z, 2, 2)
            z = 10 * self.conv2(z)
            z, s1 = self.lif1(z, s1)
            z = torch.nn.functional.max_pool2d(z, 2, 2)
            z = z.view(-1, self.features ** 2 * 50)
            z = self.fc1(z)
            z, s2 = self.lif2(z, s2)
            v, so = self.out(torch.nn.functional.relu(z), so)
            voltages[ts, :, :] = v
        return voltages


class ConvNet4(torch.nn.Module):
    def __init__(
        self, num_channels=1, feature_size=28, method="super", dtype=torch.float
    ):
        super(ConvNet4, self).__init__()
        self.features = int(((feature_size - 4) / 2 - 4) / 2)

        self.conv1 = torch.nn.Conv2d(num_channels, 32, 5, 1)
        self.conv2 = torch.nn.Conv2d(32, 64, 5, 1)
        self.fc1 = torch.nn.Linear(self.features * self.features * 64, 1024)
        self.lif0 = LIFFeedForwardCell(
            p=LIFParameters(method=method, alpha=100.0),
        )
        self.lif1 = LIFFeedForwardCell(
            p=LIFParameters(method=method, alpha=100.0),
        )
        self.lif2 = LIFFeedForwardCell(p=LIFParameters(method=method, alpha=100.0))
        self.out = LICell(1024, 10)
        self.dtype = dtype

    def forward(self, x):
        seq_length = x.shape[0]
        batch_size = x.shape[1]

        # specify the initial states
        s0 = None
        s1 = None
        s2 = None
        so = None

        voltages = torch.zeros(
            seq_length, batch_size, 10, device=x.device, dtype=self.dtype
        )

        for ts in range(seq_length):
            z = self.conv1(x[ts, :])
            z, s0 = self.lif0(z, s0)
            z = torch.nn.functional.max_pool2d(z, 2, 2)
            z = 10 * self.conv2(z)
            z, s1 = self.lif1(z, s1)
            z = torch.nn.functional.max_pool2d(z, 2, 2)
            z = z.view(-1, self.features ** 2 * 64)
            z = self.fc1(z)
            z, s2 = self.lif2(z, s2)
            v, so = self.out(torch.nn.functional.relu(z), so)
            voltages[ts, :, :] = v
        return voltages



import os
import datetime
import uuid

from absl import app
from absl import flags
from absl import logging

import torch
import torch.utils.data

from collections import namedtuple
import torchvision
import numpy as np
import matplotlib.pyplot as plt

from norse.torch import LIFParameters
from norse.torch import ConvNet, ConvNet4
from norse.torch import ConstantCurrentLIFEncoder, PoissonEncoder, SignedPoissonEncoder

FLAGS = flags.FLAGS

flags.DEFINE_bool(
    "only_first_spike", False, "Only one spike per input (latency coding)."
)
flags.DEFINE_bool("save_grads", False, "Save gradients of backward pass.")
flags.DEFINE_integer(
    "grad_save_interval", 10, "Interval for gradient saving of backward pass."
)
flags.DEFINE_string("prefix", "", "Prefix for save path to use.")
flags.DEFINE_enum(
    "encoding",
    "constant",
    ["poisson", "constant", "constant_polar", "signed_poisson", "signed_constant"],
    "Encoding to use for input",
)

flags.DEFINE_enum(
    "net", "convnet4", ["convnet", "convnet4"], "Which network architecture to use"
)
flags.DEFINE_integer("plot_interval", 10, "Interval for plotting.")
flags.DEFINE_float("input_scale", 1, "Scaling factor for input current.")
flags.DEFINE_float("current_encoder_v_th", 0.7, "v_th for constant current encoder")
flags.DEFINE_bool("learning_rate_schedule", False, "Use a learning rate schedule")
flags.DEFINE_bool("find_learning_rate", False, "Find learning rate")

flags.DEFINE_enum("device", "cpu", ["cpu", "cuda"], "Device to use by pytorch.")
flags.DEFINE_integer("epochs", 10, "Number of training episodes to do.")
flags.DEFINE_integer("seq_length", 200, "Number of timesteps to do.")
flags.DEFINE_integer("batch_size", 32, "Number of examples in one minibatch.")
flags.DEFINE_integer("hidden_size", 100, "Number of neurons in the hidden layer.")
flags.DEFINE_enum(
    "model",
    "super",
    ["super", "tanh", "circ", "logistic", "circ_dist"],
    "Model to use for training.",
)
flags.DEFINE_enum(
    "optimizer", "adam", ["adam", "sgd", "rms"], "Optimizer to use for training."
)
flags.DEFINE_float("learning_rate", 2e-3, "Learning rate to use.")
flags.DEFINE_integer(
    "log_interval", 10, "In which intervals to display learning progress."
)
flags.DEFINE_integer("model_save_interval", 50, "Save model every so many epochs.")
flags.DEFINE_boolean("save_model", True, "Save the model after training.")
flags.DEFINE_boolean("big_net", False, "Use bigger net...")
flags.DEFINE_boolean("only_output", False, "Train only the last layer...")
flags.DEFINE_boolean("do_plot", False, "Do intermediate plots")
flags.DEFINE_integer("random_seed", 1234, "Random seed to use")
flags.DEFINE_integer("start_epoch", 1, "Which epoch are we in?")
flags.DEFINE_string("resume", "", "File to resume from (if any)")
flags.DEFINE_boolean(
    "visualize_activations", False, "Should we visualize activations with visdom"
)


class PiecewiseLinear(namedtuple("PiecewiseLinear", ("batch_size", "knots", "vals"))):
    def step(self, optimizer, t):
        lr = np.interp([t], self.knots, self.vals)[0]
        for group in optimizer.param_groups:
            group["lr"] = lr / self.batch_size


def generate_poisson_trains(batch_size, num_trains, seq_length, freq):
    trains = np.random.rand(seq_length, batch_size, num_trains) < freq
    return torch.from_numpy(trains).float()


def add_luminance(images):
    return torch.cat(
        (
            images,
            torch.unsqueeze(
                0.2126 * images[0, :, :]
                + 0.7152 * images[1, :, :]
                + 0.0722 * images[2, :, :],
                0,
            ),
        ),
        0,
    )


class LIFConvNet(torch.nn.Module):
    def __init__(self, num_channels):
        super(LIFConvNet, self).__init__()

        if FLAGS.net == "convnet":
            dtype = torch.float
            self.rsnn = ConvNet(num_channels=num_channels, feature_size=32, dtype=dtype)
        elif FLAGS.net == "convnet4":
            self.rsnn = ConvNet4(num_channels=num_channels, feature_size=32)

    def forward(self, x):
        voltages = self.rsnn(x).permute(1, 0, 2)
        m, _ = torch.max(voltages, 0)
        log_p_y = torch.nn.functional.log_softmax(m, dim=1)
        return log_p_y


def train(
    model, device, train_loader, optimizer, epoch, lr_scheduler=None, writer=None
):
    model.train()
    losses = []
    train_batches = len(train_loader)
    step = train_batches * epoch

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = torch.nn.functional.nll_loss(output, target)
        loss.backward()
        if FLAGS.save_grads and batch_idx % FLAGS.grad_save_interval == 0:
            for idx, p in enumerate(model.parameters()):
                np.save(f"param-{idx}-{epoch}-{batch_idx}-grad.npy", p.grad.numpy())
                np.save(f"param-{idx}-{epoch}-{batch_idx}-data.npy", p.data.numpy())

        if lr_scheduler:
            lr_scheduler.step(optimizer, t=(epoch + batch_idx / train_batches))
        optimizer.step()
        step += 1

        if batch_idx % FLAGS.log_interval == 0:
            logging.info(
                "Train Epoch: {}/{} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    FLAGS.epochs,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )

        if step % FLAGS.log_interval == 0 and writer:
            _, argmax = torch.max(output, 1)
            accuracy = (target == argmax.squeeze()).float().mean()

            writer.add_scalar("Loss/train", loss.item(), step)
            writer.add_scalar("Accuracy/train", accuracy.item(), step)

            for tag, value in model.named_parameters():
                tag = tag.replace(".", "/")
                writer.add_histogram(tag, value.data.cpu().numpy(), step)
                writer.add_histogram(tag + "/grad", value.grad.data.cpu().numpy(), step)

        if FLAGS.do_plot and batch_idx % FLAGS.plot_interval == 0:
            ts = np.arange(0, FLAGS.seq_length)
            _, axs = plt.subplots(4, 4, figsize=(15, 10), sharex=True, sharey=True)
            axs = axs.reshape(-1)  # flatten
            for nrn in range(10):
                one_trace = model.voltages.detach().cpu().numpy()[:, 0, nrn]
                plt.sca(axs[nrn])
                plt.plot(ts, one_trace)
            plt.xlabel("Time [s]")
            plt.ylabel("Membrane Potential")
            plt.show()
        losses.append(loss.item())

    mean_loss = np.mean(losses)
    return losses, mean_loss


def test(model, device, test_loader, epoch, writer=None):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += torch.nn.functional.nll_loss(
                output, target, reduction="sum"
            ).item()  # sum up batch loss
            pred = output.argmax(
                dim=1, keepdim=True
            )  # get the index of the max log-probabilioty
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    accuracy = 100.0 * correct / len(test_loader.dataset)
    logging.info(
        f"\nTest set {FLAGS.model}: Average loss: {test_loss:.4f}, \
            Accuracy: {correct}/{len(test_loader.dataset)} ({accuracy:.0f}%)\n"
    )

    if writer:
        writer.add_scalar("Loss/test", test_loss, epoch)
        writer.add_scalar("Accuracy/test", accuracy, epoch)

    return test_loss, accuracy


def save(path, model, optimizer):
    torch.save(
        {
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
        },
        path,
    )


def load(path, model, optimizer, device):
    checkpoint = torch.load(path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    model.train(device=device)
    return model, optimizer


def main(args):
    try:
        from torch.utils.tensorboard import SummaryWriter

        writer = SummaryWriter()
    except ImportError:
        writer = None

    torch.manual_seed(FLAGS.random_seed)

    np.random.seed(FLAGS.random_seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(FLAGS.random_seed)
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True

    device = torch.device(FLAGS.device)

    constant_current_encoder = ConstantCurrentLIFEncoder(
        seq_length=FLAGS.seq_length, p=LIFParameters(v_th=FLAGS.current_encoder_v_th)
    )

    def polar_current_encoder(x):
        x_p = constant_current_encoder(2 * torch.nn.functional.relu(x))
        x_m = constant_current_encoder(2 * torch.nn.functional.relu(-x))
        return torch.cat((x_p, x_m), 1)

    def current_encoder(x):
        x = constant_current_encoder(2 * x)
        return x

    def signed_current_encoder(x):
        z = constant_current_encoder(torch.abs(x))
        return torch.sign(x) * z

    num_channels = 4

    if FLAGS.encoding == "poisson":
        encoder = PoissonEncoder(seq_length=FLAGS.seq_length, f_max=200)
    elif FLAGS.encoding == "constant":
        encoder = current_encoder
    elif FLAGS.encoding == "signed_poisson":
        encoder = SignedPoissonEncoder(seq_length=FLAGS.seq_length, f_max=200)
    elif FLAGS.encoding == "signed_constant":
        encoder = signed_current_encoder
    elif FLAGS.encoding == "constant_polar":
        encoder = polar_current_encoder
        num_channels = 2 * num_channels

    transform_train = torchvision.transforms.Compose(
        [
            torchvision.transforms.RandomCrop(32, padding=4),
            torchvision.transforms.RandomHorizontalFlip(),
            torchvision.transforms.ToTensor(),
        ]
        + [add_luminance, encoder]
    )

    transform_test = torchvision.transforms.Compose(
        [torchvision.transforms.ToTensor()] + [add_luminance, encoder]
    )

    kwargs = {"num_workers": 0, "pin_memory": True} if FLAGS.device == "cuda" else {}
    train_loader = torch.utils.data.DataLoader(
        torchvision.datasets.CIFAR10(
            root=".", train=True, download=True, transform=transform_train
        ),
        batch_size=FLAGS.batch_size,
        shuffle=True,
        **kwargs,
    )
    test_loader = torch.utils.data.DataLoader(
        torchvision.datasets.CIFAR10(root=".", train=False, transform=transform_test),
        batch_size=FLAGS.batch_size,
        **kwargs,
    )

    label = os.environ.get("SLURM_JOB_ID", str(uuid.uuid4()))
    if not FLAGS.prefix:
        rundir = f"runs/cifar10/{label}"
    else:
        rundir = f"runs/cifar10/{FLAGS.prefix}/{label}"

    os.makedirs(rundir, exist_ok=True)
    os.chdir(rundir)
    FLAGS.append_flags_into_file("flags.txt")

    model = LIFConvNet(num_channels=num_channels).to(device)

    if device == "cuda":
        model = torch.nn.DataParallel(model).to(device)

    if FLAGS.optimizer == "sgd":
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=FLAGS.learning_rate,
            momentum=0.9,
            weight_decay=5e-4 * FLAGS.batch_size,
            nesterov=True,
        )
    elif FLAGS.optimizer == "adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=FLAGS.learning_rate)
    elif FLAGS.optimizer == "rms":
        optimizer = torch.optim.RMSprop(model.parameters(), lr=FLAGS.learning_rate)

    if FLAGS.only_output:
        optimizer = torch.optim.Adam(model.out.parameters(), lr=FLAGS.learning_rate)

    if FLAGS.resume:
        if os.path.isfile(FLAGS.resume):
            checkpoint = torch.load(FLAGS.resume)
            model.load_state_dict(checkpoint["state_dict"])
            optimizer.load_state_dict(checkpoint["optimizer"])

    if FLAGS.learning_rate_schedule:
        lr_scheduler = PiecewiseLinear(
            FLAGS.batch_size, [0, 5, FLAGS.epochs], [0, 0.4, 0]
        )
    else:
        lr_scheduler = None

    training_losses = []
    mean_losses = []
    test_losses = []
    accuracies = []

    start = datetime.datetime.now()
    for epoch in range(FLAGS.start_epoch, FLAGS.start_epoch + FLAGS.epochs):
        training_loss, mean_loss = train(
            model,
            device,
            train_loader,
            optimizer,
            epoch,
            lr_scheduler=lr_scheduler,
            writer=writer,
        )
        test_loss, accuracy = test(model, device, test_loader, epoch, writer=writer)

        training_losses += training_loss
        mean_losses.append(mean_loss)
        test_losses.append(test_loss)
        accuracies.append(accuracy)

        if (epoch % FLAGS.model_save_interval == 0) and FLAGS.save_model:
            model_path = f"cifar10-{epoch}.pt"
            save(model_path, model, optimizer)

    stop = datetime.datetime.now()

    np.save("training_losses.npy", np.array(training_losses))
    np.save("mean_losses.npy", np.array(mean_losses))
    np.save("test_losses.npy", np.array(test_losses))
    np.save("accuracies.npy", np.array(accuracies))
    model_path = "cifar10-final.pt"
    save(model_path, model, optimizer)

    logging.info(f"output saved to {rundir}")
    logging.info(f"{start - stop}")
    if writer:
        writer.close()


if __name__ == "__main__":
    app.run(main)



import numpy as np
import time
import sys
from subprocess import run, STDOUT
from multiprocessing.shared_memory import ShareableList
from multiprocessing.managers import SharedMemoryManager

# pytype: disable=import-error
from pygenn.genn_model import GeNNModel
from pygenn.genn_wrapper import NO_DELAY

# pytype: enable=import-error

from .benchmark import BenchmarkParameters


def lif_feed_forward_benchmark(parameters: BenchmarkParameters):
    shared = SharedMemoryManager()
    shared.start()
    params = list(parameters._asdict().values())
    shared_list = shared.ShareableList(params)

    run(["python3", __file__, shared_list.shm.name], stderr=STDOUT)
    duration = shared_list[0]
    shared_list.shm.close()
    shared.shutdown()
    return duration


if __name__ == "__main__":
    # Assume we're running the genn benchmark and draw configs from the shared memory
    parameter_list = ShareableList(sequence=None, name=sys.argv[1])
    parameters = BenchmarkParameters(*parameter_list)

    model = GeNNModel("float", "pygenn")
    model.dT = parameters.dt
    np.random.seed(0)

    layers = []
    for i in range(parameters.batch_size):
        ones = np.ones(parameters.features)
        # Note: weights, parameters and poisson rate are magic numbers that seem to generate reasonable spike activity
        weights = np.random.rand(parameters.features, parameters.features).flatten() * 8
        model.add_neuron_population(
            f"PoissonNew{i}",
            parameters.features,
            "PoissonNew",
            {"rate": 100},
            {"timeStepToSpike": 1},
        )
        # From https://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_curr_exp
        lif_params = {
            "C": 1.0,
            "TauM": 20.0,
            "Vrest": 0.0,
            "Vreset": 0.0,
            "Vthresh": 1.0,
            "Ioffset": 0.0,
            "TauRefrac": 0.1,
        }
        lif_vars = {"V": 0.0 * ones, "RefracTime": 0.0 * ones}
        layer = model.add_neuron_population(
            f"LIF{i}", parameters.features, "LIF", lif_params, lif_vars
        )
        layers.append(layer)
        # From https://github.com/genn-team/genn/blob/master/userproject/PoissonIzh_project/model/PoissonIzh.cc#L93
        model.add_synapse_population(
            f"PL{i}",
            "DENSE_INDIVIDUALG",
            NO_DELAY,
            f"PoissonNew{i}",
            f"LIF{i}",
            "StaticPulse",
            {},
            {"g": weights},
            {},
            {},
            "DeltaCurr",
            {},
            {},
        )

    model.build()
    model.load()

    # Run simulation
    start = time.time()
    layer_spikes = []
    for _ in range(parameters.sequence_length):
        model.step_time()
        timestep_spikes = []
        # From https://github.com/neworderofjamie/pygenn_ml_tutorial/blob/master/tutorial_1.py
        for layer in layers:
            layer.pull_current_spikes_from_device()
            timestep_spikes.append(np.copy(layer.current_spikes))
        layer_spikes.append(timestep_spikes)
    end = time.time()

    parameter_list[0] = end - start
    parameter_list[1] = len(np.array(layer_spikes).flatten())
    parameter_list.shm.close()



import torch

from norse.torch.functional.lif import LIFState
from norse.torch.module.lif_mc import LIFMCCell


def test_lif_mc_cell():
    cell = LIFMCCell(2, 4)
    data = torch.randn(5, 2)
    out, s = cell(data)
    assert s.v.shape == (5, 4)
    assert s.i.shape == (5, 4)
    assert s.z.shape == (5, 4)
    assert out.shape == (5, 4)


def test_lif_mc_cell_state():
    cell = LIFMCCell(2, 4)

    input_tensor = torch.randn(5, 2)

    state = LIFState(
        z=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
        ),
        v=cell.p.v_leak
        * torch.ones(
            input_tensor.shape[0],
            cell.hidden_size,
        ),
        i=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
        ),
    )

    out, s = cell(input_tensor, state)
    assert s.v.shape == (5, 4)
    assert s.i.shape == (5, 4)
    assert s.z.shape == (5, 4)
    assert out.shape == (5, 4)


def test_lif_mc_cell_backward():
    cell = LIFMCCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)
    out.sum().backward()



"""
This module contains functional components for regularization operations on
spiking layers, where it can be desirable to regularize spikes,
membrane parameters, or other properties over time.

In this functional module, the aim is to collect some state ``s`` for each
forward step. The collection depends on the output of the layer which, by
default, simply just counts spikes. It is the job of the user to include the
regularization in an error term later.

Read more on `Wikipedia <https://en.wikipedia.org/wiki/Regularization_%28mathematics%29>`_.
"""
from typing import Any, Callable, NewType, Optional, Tuple

import torch

# An accumulator that can take some cell output, cell state, and regularization state
# and return an updated regularization state
Accumulator = NewType("Accumulator", Callable[[torch.Tensor, Any, Optional[Any]], Any])


def spike_accumulator(
    z: torch.Tensor, _: Any, state: Optional[torch.Tensor] = None
) -> int:
    """
    A spike accumulator that aggregates spikes and returns the total sum as an integer.

    Parameters:
        z (torch.Tensor): Spikes from some cell
        s (Any): Cell state
        state (Optional[int]): The regularization state to be aggregated to. Defaults to 0.

    Returns:
        A new RegularizationState containing the aggregated data
    """
    if state is None:
        state = 0
    return state + z.sum()


def voltage_accumulator(
    z: torch.Tensor, s: Any, state: Optional[torch.Tensor] = None
) -> int:
    """
    A spike accumulator that aggregates membrane potentials over time. Requires that the
    input state ``s`` has a ``v`` property (for voltage).

    Parameters:
        z (torch.Tensor): Spikes from some cell
        s (Any): Cell state
        state (Optional[torch.Tensor]): The regularization state to be aggregated to.
        Defaults to zeros for all entries.

    Returns:
        A new RegularizationState containing the aggregated data
    """
    if state is None:
        state = torch.zeros_like(s.v)
    return state + s.v


def regularize_step(
    z: torch.Tensor,
    s: Any,
    accumulator: Accumulator = spike_accumulator,
    state: Any = None,
) -> Tuple[torch.Tensor, Any]:
    """
    Takes one step for a regularizer that aggregates some information (based on the
    spike_accumulator function), which is pushed forward and returned for future
    inclusion in an error term.

    Parameters:
        z (torch.Tensor): Spikes from a cell
        s (Any): Neuron state
        state (Optional[Any]): The regularization state to be aggregated. Typically some numerical value like a count. Defaults to None

    Return:
        A tuple of (spikes, regularizer state)
    """
    if state is None:
        return z, accumulator(z, s, None)
    return z, accumulator(z, s, state)



from .conv import ConvNet, ConvNet4
from .vgg import VGG, vgg11, vgg11_bn, vgg13, vgg13_bn, vgg16, vgg16_bn, vgg19, vgg19_bn
from .mobilenet import MobileNetV2, mobilenet_v2

__all__ = [
    "ConvNet",
    "ConvNet4",
    "VGG",
    "vgg11",
    "vgg11_bn",
    "vgg13",
    "vgg13_bn",
    "vgg16",
    "vgg16_bn",
    "vgg19",
    "vgg19",
    "vgg19_bn",
    "MobileNetV2",
    "mobilenet_v2",
]



import torch

from norse.torch.module.leaky_integrator import LICell, LIFeedForwardCell, LIState


def test_li_cell():
    cell = LICell(2, 4)
    data = torch.randn(5, 2)
    out, s = cell(data)

    for x in s:
        assert x.shape == (5, 4)
    assert out.shape == (5, 4)


def test_li_cell_state():
    cell = LICell(2, 4)
    data = torch.randn(5, 2)
    out, s = cell(data, LIState(torch.ones(5, 4), torch.ones(5, 4)))

    for x in s:
        assert x.shape == (5, 4)
    assert out.shape == (5, 4)


def test_cell_backward():
    model = LICell(12, 1)
    data = torch.ones(100, 12)
    out, _ = model(data)
    loss = out.sum()
    loss.backward()


def test_li_feedforward_cell():
    layer = LIFeedForwardCell()
    data = torch.randn(10, 2, 4)
    out, _ = layer(data)

    assert out.shape == (10, 2, 4)


def test_li_feedforward_cell_state():
    layer = LIFeedForwardCell()
    data = torch.randn(2, 4)
    out, s = layer(data, LIState(torch.ones(2, 4), torch.ones(2, 4)))

    for x in s:
        assert x.shape == (2, 4)
    assert out.shape == (2, 4)


def test_ff_backward():
    model = LIFeedForwardCell()
    data = torch.ones(10, 12, 1)
    out, _ = model(data)
    loss = out.sum()
    loss.backward()



import torch

from norse.torch.functional.leaky_integrator import (
    LIState,
    li_feed_forward_step,
    li_step,
)


def test_lif_step():
    x = torch.ones(20)
    s = LIState(v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.randn(10, 20).float()

    for _ in range(100):
        _, s = li_step(x, s, input_weights)


def test_lif_feed_forward_step():
    x = torch.ones(10)
    s = LIState(v=torch.zeros(10), i=torch.zeros(10))

    for _ in range(100):
        _, s = li_feed_forward_step(x, s)



import time
import torch

# pytype: disable=import-error
from bindsnet.network import Network
from bindsnet.network.topology import Connection
from bindsnet.network.nodes import LIFNodes, Input
from bindsnet.encoding import poisson
from bindsnet.network.monitors import Monitor

# pytype: enable=import-error

from .benchmark import BenchmarkParameters


class BindsNetModule(torch.nn.Module):
    def __init__(self, parameters: BenchmarkParameters):
        super(BindsNetModule, self).__init__()
        network = Network(batch_size=parameters.batch_size, dt=parameters.dt)
        lif_nodes = LIFNodes(n=parameters.features)
        monitor = Monitor(
            obj=lif_nodes, state_vars=("s"), time=parameters.sequence_length
        )
        network.add_layer(Input(n=parameters.features), name="Input")
        network.add_layer(lif_nodes, name="Neurons")
        network.add_connection(
            Connection(
                source=network.layers["Input"], target=network.layers["Neurons"]
            ),
            source="Input",
            target="Neurons",
        )
        network.add_monitor(monitor, "Monitor")
        network.to(parameters.device)

        self.parameters = parameters
        self.network = network
        self.monitor = monitor

    def forward(self, input_data):
        self.network.run(
            inputs=input_data, time=self.parameters.dt * self.parameters.sequence_length
        )
        return self.monitor.get("s")


def lif_feed_forward_benchmark(parameters: BenchmarkParameters):
    with torch.no_grad():
        network = BindsNetModule(parameters)
        input_data = {
            "Input": poisson(
                datum=torch.rand(
                    (parameters.batch_size, parameters.features),
                    device=parameters.device,
                ).contiguous(),
                time=parameters.sequence_length,
                device=parameters.device,
            )
        }

        start = time.time()
        network(input_data)
        end = time.time()

        duration = end - start
        return duration



import importlib
from absl import app
from absl import flags

FLAGS = flags.FLAGS
flags.DEFINE_enum("task", "mnist", ["cifar", "gym", "mnist"], "Task to run.")


def main(argv):
    task_module = importlib.import_module(
        f"norse.examples.run_{FLAGS.task}", package="."
    )
    app.run(task_module.main)


if __name__ == "__main__":
    app.run(main)



import torch

from norse.torch.functional.lif import LIFState, LIFFeedForwardState
from norse.torch.functional.lif_refrac import (
    LIFRefracState,
    LIFRefracFeedForwardState,
    lif_refrac_feed_forward_step,
    lif_refrac_step,
)


def test_lif_refrac_step():
    x = torch.ones(20)
    s = LIFRefracState(
        lif=LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10)),
        rho=torch.zeros(10),
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lif_refrac_step(x, s, input_weights, recurrent_weights)


def test_lif_refrac_feed_forward_step():
    x = torch.ones(10)
    s = LIFRefracFeedForwardState(
        lif=LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10)),
        rho=torch.zeros(10),
    )

    for _ in range(100):
        _, s = lif_refrac_feed_forward_step(x, s)



import inspect
import torch
from typing import Union


class SequentialState(torch.nn.Sequential):
    """
    A sequential model that works exactly like PyTorch's ``Sequential`` with the
    addition that it handles neuron states.

    Example:
        >>> import torch
        >>> from norse.torch.module.lift import LIFT
        >>> from norse.torch.module.sequential import SequentialState
        >>> from norse.torch.module.lif import LIFLayer
        >>> data = torch.ones(1, 16, 8, 4)     # Single timestep
        >>> model = SequentialState(
        >>>   Lift(torch.nn.Conv2d(16, 8, 3)), # (1, 8, 6, 2)
        >>>   torch.nn.Flatten(2),             # (1, 8, 12)
        >>>   LIFLayer(12, 6),                 # (1, 8, 6)
        >>>   LIFLayer(6, 1)                   # (1, 8, 1)
        >>> )
        >>> model(data)

    Example with recurrent layers:
        >>> import torch
        >>> from norse.torch.module.lift import Lift
        >>> from norse.torch.module.sequential import SequentialState
        >>> from norse.torch.module.lif import LIFCell, LIFLayer
        >>> from norse.torch.module.lsnn import LSNNCell, LSNNLayer
        >>> data = torch.ones(1, 16, 8, 4)     # Single timestep
        >>> model = SequentialState(
        >>>   Lift(torch.nn.Conv2d(16, 8, 3)), # (1, 8, 6, 2)
        >>>   torch.nn.Flatten(2),             # (1, 8, 12)
        >>>   LSNNLayer(LSNNCell, 12, 6),      # (1, 8, 6)
        >>>   torch.nn.RNN(6, 4, 2),           # (1, 6, 4) with 2 recurrent layers
        >>>   LIFLayer(LIFCell,4, 1)           # (1, 4, 1)
        >>> )
        >>> model(data)

    """

    def __init__(self, *args: torch.nn.Module):
        super(SequentialState, self).__init__()
        self.stateful_layers = []
        for idx, module in enumerate(args):
            self.add_module(str(idx), module)
            # Identify all the stateful layers
            signature = inspect.signature(module.forward)
            self.stateful_layers.append(
                "state" in signature.parameters or isinstance(module, torch.nn.RNNBase)
            )

    def forward(self, input_tensor: torch.Tensor, state: Union[list, None] = None):
        """
        Feeds the input to the modules with the given state-list.
        If the state is None, the initial state is set to None for each of the modules.

        Parameters:
            input_tensor: The input tensor too feed into the first module
            state: Either a list of states for each module or None. If None, the modules
                   will initialise their own default state

        Returns:
            A tuple of (output tensor, state list)
        """
        state = [None] * len(self) if state is None else state
        for index, module in enumerate(self):
            if self.stateful_layers[index]:
                input_tensor, s = module(input_tensor, state[index])
                state[index] = s
            else:
                input_tensor = module(input_tensor)
        return input_tensor, state



from typing import NamedTuple, Tuple

import torch
import torch.jit

from norse.torch.functional.lif import LIFState, LIFParameters, lif_step
from norse.torch.functional.correlation_sensor import (
    CorrelationSensorState,
    CorrelationSensorParameters,
    correlation_sensor_step,
)


class LIFCorrelationState(NamedTuple):
    lif_state: LIFState
    input_correlation_state: CorrelationSensorState
    recurrent_correlation_state: CorrelationSensorState


class LIFCorrelationParameters(NamedTuple):
    lif_parameters: LIFParameters = LIFParameters()
    input_correlation_parameters: CorrelationSensorParameters = (
        CorrelationSensorParameters()
    )
    recurrent_correlation_parameters: CorrelationSensorParameters = (
        CorrelationSensorParameters()
    )


def lif_correlation_step(
    input_tensor: torch.Tensor,
    state: LIFCorrelationState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFCorrelationParameters = LIFCorrelationParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFCorrelationState]:
    z_new, s_new = lif_step(
        input_tensor,
        state.lif_state,
        input_weights,
        recurrent_weights,
        p.lif_parameters,
        dt,
    )

    input_correlation_state_new = correlation_sensor_step(
        z_pre=input_tensor,
        z_post=z_new,
        state=state.input_correlation_state,
        p=p.input_correlation_parameters,
        dt=dt,
    )

    recurrent_correlation_state_new = correlation_sensor_step(
        z_pre=state.lif_state.z,
        z_post=z_new,
        state=state.recurrent_correlation_state,
        p=p.recurrent_correlation_parameters,
        dt=dt,
    )
    return (
        z_new,
        LIFCorrelationState(
            lif_state=s_new,
            input_correlation_state=input_correlation_state_new,
            recurrent_correlation_state=recurrent_correlation_state_new,
        ),
    )



import setuptools
from setuptools import setup

from os import path

from torch.utils.cpp_extension import BuildExtension, CppExtension

pwd = path.abspath(path.dirname(__file__))

with open(path.join(pwd, "requirements.txt")) as fp:
    install_requires = fp.read()

with open(path.join(pwd, "README.md"), encoding="utf-8") as fp:
    readme_text = fp.read()

setup(
    install_requires=install_requires,
    name="norse",
    version="0.0.5",
    description="A library for deep learning with spiking neural networks",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    url="http://github.com/norse/norse",
    author="Christian Pehle",
    author_email="christian.pehle@gmail.com",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="machine learning spiking neural networks",
    ext_modules=[
        CppExtension("norse_op", ["norse/csrc/op.cpp"], extra_compile_args=["-O3"])
    ],
    cmdclass={"build_ext": BuildExtension.with_options(no_python_abi_suffix=True)},
)



import torch

from norse.torch.functional.tsodyks_makram import (
    stp_step,
    TsodyksMakramState,
    TsodyksMakramParameters,
)


def example(p):
    dt = 0.001
    z = torch.zeros(1000)
    z[::100] = 1.0
    z[0:10] = 0.0

    s = TsodyksMakramState(x=1.0, u=0.0)
    i = 0.0
    xs = []
    us = []
    current = []

    for ts in range(1000):
        x, s = stp_step(z[ts], s, p)
        di = -p.tau_s_inv * i
        i = i + dt * di + x
        xs += [s.x]
        us += [s.u]
        current += [i]

    xs = torch.stack(xs)
    us = torch.stack(us)
    current = torch.stack(current)

    return xs, us, current


def test_depressing():
    p = TsodyksMakramParameters(
        tau_f_inv=1 / (50.0e-3),
        tau_s_inv=1 / (20.0e-3),
        tau_d_inv=1 / (750.0e-3),
        U=0.45,
    )
    _, _, current = example(p)
    assert torch.allclose(current[100], torch.as_tensor(p.U))
    assert current[500:1000].max() < current[0:500].max()


def test_facilitating():
    p = TsodyksMakramParameters(
        tau_f_inv=1 / (750.0e-3),
        tau_s_inv=1 / (20.0e-3),
        tau_d_inv=1 / (50.0e-3),
        U=0.15,
    )
    _, _, current = example(p)
    assert torch.allclose(current[100], torch.as_tensor(p.U))
    assert current[500:1000].max() > current[0:500].max()



import torch
import torch.jit

from typing import NamedTuple, Tuple


class LIState(NamedTuple):
    """State of a leaky-integrator

    Parameters:
        v (torch.Tensor): membrane voltage
        i (torch.Tensor): input current
    """

    v: torch.Tensor
    i: torch.Tensor


class LIParameters(NamedTuple):
    """Parameters of a leaky integrator

    Parameters:
        tau_syn_inv (torch.Tensor): inverse synaptic time constant
        tau_mem_inv (torch.Tensor): inverse membrane time constant
        v_leak (torch.Tensor): leak potential
    """

    tau_syn_inv: torch.Tensor = torch.as_tensor(1.0 / 5e-3)
    tau_mem_inv: torch.Tensor = torch.as_tensor(1.0 / 1e-2)
    v_leak: torch.Tensor = torch.as_tensor(0.0)


def li_step(
    input_tensor: torch.Tensor,
    state: LIState,
    input_weights: torch.Tensor,
    p: LIParameters = LIParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIState]:
    r"""Single euler integration step of a leaky-integrator.
    More specifically it implements a discretized version of the ODE

    .. math::

        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}


    and transition equations

    .. math::
        i = i + w i_{\text{in}}

    Parameters:
        input_tensor (torch.Tensor); Input spikes
        s (LIState): state of the leaky integrator
        input_weights (torch.Tensor): weights for incoming spikes
        p (LIParameters): parameters of the leaky integrator
        dt (float): integration timestep to use
    """

    # compute voltage updates
    dv = dt * p.tau_mem_inv * ((p.v_leak - state.v) + state.i)
    v_new = state.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute current jumps
    i_new = i_decayed + torch.nn.functional.linear(input_tensor, input_weights)
    return v_new, LIState(v_new, i_new)


# @torch.jit.script
def li_feed_forward_step(
    input_tensor: torch.Tensor,
    state: LIState,
    p: LIParameters = LIParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIState]:
    # compute voltage updates
    dv = dt * p.tau_mem_inv * ((p.v_leak - state.v) + state.i)
    v_new = state.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute current jumps
    i_new = i_decayed + input_tensor
    return v_new, LIState(v_new, i_new)



"""
Test for the stateful encoder module
"""

import torch

from ..lif import LIFLayer
from norse.torch.module.encode import (
    PopulationEncoder,
    ConstantCurrentLIFEncoder,
    SpikeLatencyEncoder,
    PoissonEncoder,
    SignedPoissonEncoder,
    SpikeLatencyLIFEncoder,
)

# Fixes a linting error:
# pylint: disable=E1102


def test_encode_population():
    data = torch.as_tensor([0, 0.5, 1])
    out_features = 3
    actual = PopulationEncoder(out_features).forward(data)
    expected = torch.as_tensor(
        [
            [1.0000, 0.8824969, 0.6065307],
            [0.8824969, 1.0000, 0.8824969],
            [0.6065307, 0.8824969, 1.0000],
        ]
    )
    assert torch.allclose(actual, expected)


def test_constant_current_lif_encode():
    data = torch.as_tensor([0, 0, 0, 0])
    z = ConstantCurrentLIFEncoder(2).forward(data)
    assert torch.equal(z, torch.zeros((2, 4)))


def test_spike_latency_encode():
    data = torch.as_tensor([[0, 100, 100], [100, 100, 100]])
    encoder = torch.nn.Sequential(ConstantCurrentLIFEncoder(2), SpikeLatencyEncoder())
    actual = encoder(data)
    expected = torch.zeros((2, 2, 3))
    expected[0] = torch.as_tensor([[0, 1, 1], [1, 1, 1]])
    assert torch.equal(actual, expected)


def test_signed_poisson_encoder():
    seq_length = 100
    encoder = SignedPoissonEncoder(seq_length=seq_length)
    x = torch.randn(10, 10)
    out = encoder(x)
    assert out.shape == torch.Size([seq_length, 10, 10])


def test_poisson_encoder():
    seq_length = 100
    encoder = PoissonEncoder(seq_length=seq_length)
    x = torch.randn(10, 10)
    out = encoder(x)
    assert out.shape == torch.Size([seq_length, 10, 10])


def test_spike_latency_lif_encoder():
    seq_length = 100
    encoder = SpikeLatencyLIFEncoder(seq_length=seq_length)
    x = torch.randn(10, 10)
    out = encoder(x)
    assert out.shape == torch.Size([seq_length, 10, 10])


def test_spike_latency_lif_encoder_backward():
    timesteps = 20
    data = torch.randn(7, 5) + 10
    encoder = SpikeLatencyLIFEncoder(timesteps)
    layer = LIFLayer(5, 2)
    encoded = encoder(data)
    out, _ = layer(encoded)
    out.sum().backward()


def test_spike_latency_encode_max_spikes():
    encoder = torch.nn.Sequential(
        ConstantCurrentLIFEncoder(seq_length=128), SpikeLatencyEncoder()
    )
    spikes = encoder(1.1 * torch.ones(10))
    assert torch.sum(spikes).data == 10


def test_spike_latency_encode_backward():
    torch.autograd.set_detect_anomaly(True)
    timesteps = 20
    data = torch.randn(7, 5) + 10
    encoder = torch.nn.Sequential(
        ConstantCurrentLIFEncoder(seq_length=timesteps), SpikeLatencyEncoder()
    )
    layer = LIFLayer(5, 2)
    encoded = encoder(data)
    out, _ = layer(encoded)
    out = out.sum()
    out.backward()


def test_spike_latency_encode_chain():
    data = torch.randn(7, 5) + 10
    encoder = torch.nn.Sequential(ConstantCurrentLIFEncoder(2), SpikeLatencyEncoder())
    encoder(data)



"""
Tests that the training of Norse modules stays intact, for instance
that gradients are properly propagated
"""

import torch

from norse.torch.module.encode import PoissonEncoder
from norse.torch.module.lif import LIFCell


class SNNetwork(torch.nn.Module):
    def __init__(self):
        super(SNNetwork, self).__init__()
        self.encoder = PoissonEncoder(10, f_max=1000)
        self.l0 = LIFCell(12, 6)
        self.l1 = LIFCell(6, 1)
        self.s0 = self.s1 = None

    def forward(self, input_tensor):
        spike_ts = self.encoder(input_tensor)
        spikes = None
        for spikes in spike_ts:
            spikes, self.s0 = self.l0(spikes, self.s0)
            spikes, self.s1 = self.l1(spikes, self.s1)
        return spikes


def test_optimize_model():
    model = SNNetwork()
    optimizer = torch.optim.Adam(model.parameters(), lr=1)
    optimizer.zero_grad()
    input_weights = model.l0.input_weights.clone()
    recurrent_weights = model.l1.recurrent_weights.clone()
    data = torch.ones(1, 12)
    out = model(data)
    loss = out.sum()
    loss.backward()
    optimizer.step()
    assert not torch.all(torch.eq(input_weights, model.l0.input_weights))
    assert not torch.all(torch.eq(recurrent_weights, model.l0.recurrent_weights))



import torch
import torch.jit
import numpy as np

import norse
from norse.torch.functional.heaviside import heaviside

if getattr(norse, "IS_OPS_LOADED"):
    superspike_fn = torch.ops.norse_op.superfun
else:  #  pragma: no cover
    from .superspike import super_fn

    superspike_fn = super_fn


class HeaviErfc(torch.autograd.Function):
    r"""Approximation of the heaviside step function as

    .. math::
        h(x,k) = \frac{1}{2} + \frac{1}{2} \text{erfc}(k x)

    where erfc is the error function.
    """

    @staticmethod
    def forward(ctx, x, k):
        ctx.save_for_backward(x)
        ctx.k = k
        return heaviside(x)  # 0 + 0.5 * torch.erfc(k * x)

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        derfc = (2 * torch.exp(-(ctx.k * x).pow(2))) / (torch.as_tensor(np.pi).sqrt())
        return derfc * dy, None


@torch.jit.ignore
def heavi_erfc_fn(x: torch.Tensor, k: float):
    return HeaviErfc.apply(x, k)


class HeaviTanh(torch.autograd.Function):
    r"""Approximation of the heaviside step function as

    .. math::
        h(x,k) = \frac{1}{2} + \frac{1}{2} \text{tanh}(k x)
    """

    @staticmethod
    def forward(ctx, x, k):
        ctx.save_for_backward(x)
        ctx.k = k
        return heaviside(x)  # 0.5 + 0.5 * torch.tanh(k * x)

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        dtanh = 1 - (x * ctx.k).tanh().pow(2)
        return dy * dtanh, None


@torch.jit.ignore
def heavi_tanh_fn(x: torch.Tensor, k: float):
    return HeaviTanh.apply(x, k)


class Logistic(torch.autograd.Function):
    r"""Probalistic approximation of the heaviside step function as

    .. math::
        z \sim p(\frac{1}{2} + \frac{1}{2} \text{tanh}(k x))
    """

    @staticmethod
    def forward(ctx, x, k):
        ctx.k = k
        ctx.save_for_backward(x)
        p = 0.5 + 0.5 * torch.tanh(k * x)
        return torch.distributions.bernoulli.Bernoulli(probs=p).sample()

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        dtanh = 1 - (x * ctx.k).tanh().pow(2)
        return dy * dtanh, None


@torch.jit.ignore
def logistic_fn(x: torch.Tensor, k: float):
    return Logistic.apply(x, k)


class HeaviCirc(torch.autograd.Function):
    r"""Approximation of the heaviside step function as

    .. math::
        h(x,\alpha) = \frac{1}{2} + \frac{1}{2} \
        \frac{x}{(x^2 + \alpha^2)^{1/2}}
    """

    @staticmethod
    def forward(ctx, x, alpha):
        ctx.save_for_backward(x)
        ctx.alpha = alpha
        return heaviside(x)  # 0.5 + 0.5 * (x / (x.pow(2) + alpha ** 2).sqrt())

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        alpha = ctx.alpha

        return (
            dy
            * (
                -(x.pow(2) / (2 * (alpha ** 2 + x.pow(2)).pow(1.5)))
                + 1 / (2 * (alpha ** 2 + x.pow(2)).sqrt())
            )
            * 2
            * alpha,
            None,
        )


@torch.jit.ignore
def heavi_circ_fn(x: torch.Tensor, k: float):
    return HeaviCirc.apply(x, k)


class CircDist(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, alpha):
        ctx.save_for_backward(x)
        ctx.alpha = alpha

        return torch.distributions.bernoulli.Bernoulli(
            0.5 + 0.5 * (x / (x.pow(2) + alpha ** 2).sqrt())
        ).sample()

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        alpha = ctx.alpha
        return (
            dy
            * (
                -(x.pow(2) / (2 * (alpha ** 2 + x.pow(2)).pow(1.5)))
                + 1 / (2 * (alpha ** 2 + x.pow(2)).sqrt())
            )
            * 2
            * alpha,
            None,
        )


@torch.jit.ignore
def circ_dist_fn(x: torch.Tensor, k: float):
    return CircDist.apply(x, k)


class HeaviTent(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, alpha):
        ctx.alpha = alpha
        ctx.save_for_backward(x)
        return heaviside(x)

    @staticmethod
    def backward(ctx, dy):
        (x,) = ctx.saved_tensors
        alpha = ctx.alpha
        return torch.relu(1 - torch.abs(x)) * alpha * dy, None


@torch.jit.ignore
def heavi_tent_fn(x: torch.Tensor, k: float):
    return HeaviTent.apply(x, k)


def threshold(x: torch.Tensor, method: str, alpha: float) -> torch.Tensor:
    if method == "heaviside":
        return heaviside(x)
    elif method == "super":
        return superspike_fn(x, torch.as_tensor(alpha))
    elif method == "tanh":
        return heavi_tanh_fn(x, alpha)
    elif method == "tent":
        return heavi_tent_fn(x, alpha)
    elif method == "circ":
        return heavi_circ_fn(x, alpha)
    elif method == "heavi_erfc":
        return heavi_erfc_fn(x, alpha)
    else:
        raise ValueError(
            f"Attempted to apply threshold function {method}, but no such "
            + "function exist. We currently support heaviside, super, "
            + "tanh, tent, circ, and heavi_erfc."
        )


def sign(x: torch.Tensor, method: str, alpha: float) -> torch.Tensor:
    return 2 * threshold(x, method, alpha) - 1



"""This module provides a wrapper around the torchaudio.SPEECHCOMMANDS dataset.
"""
import torch

# pytype: disable=import-error
import torchaudio

# pytype: enable=import-error
import os
import glob


def generate_background_noise(speech_commands):
    """Split the background noise provided by the dataset in 1 second chunks.

    Parameters:
        speech_commands (torch.utils.data.Dataset): Speech Command dataset as defined by torchaudio.
    """
    background_noise = glob.glob(
        os.path.join(speech_commands._path, "_background_noise_", "*.wav")
    )
    os.makedirs(os.path.join(speech_commands._path, "background"), exist_ok=True)

    for file in background_noise:
        waveform, sample_rate = torchaudio.load(file)
        background_waveforms = torch.split(waveform, sample_rate, dim=1)[:-1]

        for idx, background_waveform in enumerate(background_waveforms):
            torchaudio.save(
                os.path.join(
                    speech_commands._path,
                    "background",
                    f"{hash(waveform)}_nohash_{idx}.wav",
                ),
                background_waveform,
                sample_rate=sample_rate,
            )


def prepare_dataset(speech_commands):
    """Prepare the speech command dataset by splitting it into training, validation
    and testing according to the method described in "Speech Commands: A Dataset for Limited-Vocabulary Speech Recognition"
    (https://arxiv.org/abs/1804.03209). It is meant to be used with the SPEECHCOMMANDS
    dataset provided by torchaudio.

    .. code-block:: python
        speech_commands = torchaudio.datasets.SPEECHCOMMANDS(root='.', download=True)
        train_sc, valid_sc, test_sc = prepare_dataset(speech_commands)

    Parameters:
        speech_commands (torch.utils.data.Dataset): Speech Commands dataset as defined by torchaudio.
    """
    with open(os.path.join(speech_commands._path, "validation_list.txt")) as f:
        text = f.read()
        valid_filelist = text.split("\n")
        valid_filelist = [
            os.path.join(speech_commands._path, p) for p in valid_filelist
        ]

    with open(os.path.join(speech_commands._path, "testing_list.txt")) as f:
        text = f.read()
        test_filelist = text.split("\n")
        test_filelist = [os.path.join(speech_commands._path, p) for p in test_filelist]

    train_sc = torchaudio.datasets.SPEECHCOMMANDS(root=".")
    train_sc._walker = list(
        filter(
            lambda w: w not in valid_filelist and w not in test_filelist,
            train_sc._walker,
        )
    )

    valid_sc = torchaudio.datasets.SPEECHCOMMANDS(root=".")
    valid_sc._walker = list(filter(lambda w: w in valid_filelist, valid_sc._walker))

    test_sc = torchaudio.datasets.SPEECHCOMMANDS(root=".")
    test_sc._walker = list(filter(lambda w: w in valid_filelist, valid_sc._walker))

    # no generate the background noise category
    generate_background_noise(speech_commands)

    ratios = [
        len(train_sc._walker) / len(speech_commands._walker),
        len(valid_sc._walker) / len(speech_commands._walker),
        len(test_filelist) / len(speech_commands._walker),
    ]

    files = glob.glob(os.path.join(speech_commands._path, "background/*.wav"))
    train_idx = int(ratios[0] * len(files))
    valid_idx = train_idx + int(ratios[1] * len(files))
    test_idx = valid_idx + int(ratios[2] * len(files))
    training = files[:train_idx]
    valid = files[train_idx:valid_idx]
    test = files[valid_idx : (test_idx + 1)]

    train_sc._walker += training
    test_sc._walker += test
    valid_sc._walker += valid

    return train_sc, valid_sc, test_sc


def label_to_index(dataset):
    """Generate integer labels for each of the classes in the speech commands dataset.

    Parameters:
        speech_commands (torch.utils.data.Dataset): Speech Commands dataset as defined by torchaudio.
    """
    known = {
        "yes",
        "no",
        "up",
        "down",
        "left",
        "right",
        "on",
        "off",
        "stop",
        "go",
        "background",
    }
    all_labels = set()

    for (_, _, label, _, _) in dataset:
        all_labels.add(label)

    unknown = all_labels - known
    label_to_index = {
        **{key: idx for idx, key in enumerate(known)},
        **{key: 12 for key in unknown},
    }
    return label_to_index


class SpeechCommands(torch.utils.data.Dataset):
    """Speech Commands dataset as described in "Speech Commands: A Dataset for Limited-Vocabulary Speech Recognition"
    (https://arxiv.org/abs/1804.03209). This is meant as a wrapper around the corresponding
    SPEECHCOMMANDS dataset defined in torchaudio.

    .. code-block:: python
        speech_commands = torchaudio.datasets.SPEECHCOMMANDS(root='.', download=True)
        train_sc, valid_sc, test_sc = prepare_dataset(speech_commands)

        train_transform = torch.nn.Sequential(
            torchaudio.transforms.MelSpectrogram(sample_rate=16000, n_mels=81),
            torchaudio.transforms.FrequencyMasking(freq_mask_param=15),
            torchaudio.transforms.TimeMasking(time_mask_param=35)
        )

        valid_transform = torch.nn.Sequential(
            torchaudio.transforms.MelSpectrogram(sample_rate=16000, n_mels=81),
        )

        train_speech_commands = SpeechCommands(dataset=train_sc, transform=train_transform)
        valid_speech_commands = SpeechCommands(dataset=valid_sc, transform=valid_transform)

    Parameters:
        speech_commands (torch.utils.data.Dataset): Speech Commands dataset as defined by torchaudio.
        transform (torch.nn.Module): Sequence of transformations to apply to waveform data.
    """

    def __init__(self, dataset, transform):
        self.dataset = dataset
        self.transform = transform
        self.label_to_index = label_to_index(dataset)

    def __getitem__(self, index):
        (waveform, sample_rate, label, _, _) = self.dataset[index]

        padding = int((sample_rate - waveform.shape[1]))
        waveform = torch.nn.functional.pad(waveform, (0, padding))
        spec = self.transform(waveform)
        idx = self.label_to_index[label]

        return spec, idx

    def __len__(self):
        return len(self.dataset)



from typing import NamedTuple, Tuple

import torch

from norse.torch.functional.threshold import threshold


class LSNNParameters(NamedTuple):
    r"""Parameters of an LSNN neuron

    Parameters:
        tau_syn_inv (torch.Tensor): inverse synaptic time
                                    constant (:math:`1/\tau_\text{syn}`)
        tau_mem_inv (torch.Tensor): inverse membrane time
                                    constant (:math:`1/\tau_\text{mem}`)
        tau_adapt_inv (torch.Tensor): inverse adaptation time
                                      constant (:math:`1/\tau_b`)
        v_leak (torch.Tensor): leak potential
        v_th (torch.Tensor): threshold potential
        v_reset (torch.Tensor): reset potential
        beta (torch.Tensor): adaptation constant
    """

    tau_syn_inv: torch.Tensor = torch.as_tensor(1.0 / 5e-3)
    tau_mem_inv: torch.Tensor = torch.as_tensor(1.0 / 1e-2)
    tau_adapt_inv: torch.Tensor = torch.as_tensor(1.0 / 700)
    v_leak: torch.Tensor = torch.as_tensor(0.0)
    v_th: torch.Tensor = torch.as_tensor(1.0)
    v_reset: torch.Tensor = torch.as_tensor(0.0)
    beta: torch.Tensor = torch.as_tensor(1.8)
    method: str = "super"
    alpha: float = 100.0


class LSNNState(NamedTuple):
    """State of an LSNN neuron

    Parameters:
        z (torch.Tensor): recurrent spikes
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
        b (torch.Tensor): threshold adaptation
    """

    z: torch.Tensor
    v: torch.Tensor
    i: torch.Tensor
    b: torch.Tensor

    def clone(self):
        return LSNNState(self.z.clone(), self.v.clone(), self.i.clone(), self.b.clone())

    def detach(self):
        return LSNNState(
            self.z.detach(), self.v.detach(), self.i.detach(), self.b.detach()
        )


def lsnn_step(
    input_tensor: torch.Tensor,
    state: LSNNState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LSNNParameters = LSNNParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LSNNState]:
    r"""Euler integration step for LIF Neuron with threshold adaptation
    More specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i \\
            \dot{b} &= -1/\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}} + b)

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + w_{\text{input}} z_{\text{in}} \\
            i &= i + w_{\text{rec}} z_{\text{rec}} \\
            b &= b + \beta z
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the recurrent
    and input spikes respectively.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LSNNState): current state of the lsnn unit
        input_weights (torch.Tensor): synaptic weights for input spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """
    # compute voltage decay
    dv = dt * p.tau_mem_inv * ((p.v_leak - state.v) + state.i)
    v_decayed = state.v + dv

    # compute current decay
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute threshold adaptation update
    db = dt * p.tau_adapt_inv * (p.v_th - state.b)
    b_decayed = state.b + db

    # compute new spikes
    z_new = threshold(v_decayed - b_decayed, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = (
        i_decayed
        + torch.nn.functional.linear(input_tensor, input_weights)
        + torch.nn.functional.linear(state.z, recurrent_weights)
    )

    b_new = b_decayed + z_new * p.tau_adapt_inv * p.beta
    return z_new, LSNNState(z_new, v_new, i_new, b_new)


def ada_lif_step(
    input_tensor: torch.Tensor,
    state: LSNNState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LSNNParameters = LSNNParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LSNNState]:
    r"""Euler integration step for LIF Neuron with adaptation. More specifically
    it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + b + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i \\
            \dot{b} &= -1/\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\
            i &= i + w_{\text{input}} z_{\\text{in}} \\
            i &= i + w_{\text{rec}} z_{\\text{rec}} \\
            b &= b + \beta z
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the recurrent
    and input spikes respectively.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LSNNState): current state of the lsnn unit
        input_weights (torch.Tensor): synaptic weights for input spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """
    di = -dt * p.tau_syn_inv * state.i
    i = state.i + di
    i = i + torch.nn.functional.linear(input_tensor, input_weights)
    i = i + torch.nn.functional.linear(state.z, recurrent_weights)
    dv = dt * p.tau_mem_inv * ((p.v_leak - state.v) + state.i - state.b)
    v = state.v + dv
    db = -dt * p.tau_adapt_inv * state.b
    b = state.b + db
    z_new = threshold(v - p.v_th, p.method, p.alpha)
    v = v - z_new * (p.v_th - p.v_reset)
    b = b + z_new * p.tau_adapt_inv * p.beta
    return z_new, LSNNState(z_new, v, i, b)


class LSNNFeedForwardState(NamedTuple):
    """Integration state kept for a lsnn module

    Parameters:
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
        b (torch.Tensor): threshold adaptation
    """

    v: torch.Tensor
    i: torch.Tensor
    b: torch.Tensor


def lsnn_feed_forward_step(
    input_tensor: torch.Tensor,
    state: LSNNFeedForwardState,
    p: LSNNParameters = LSNNParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LSNNFeedForwardState]:
    r"""Euler integration step for LIF Neuron with threshold adaptation.
    More specifically it implements one integration step of the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i \\
            \dot{b} &= -1/\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}} + b)

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + \text{input} \\
            b &= b + \beta z
        \end{align*}

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LSNNFeedForwardState): current state of the lsnn unit
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """
    # compute voltage updates
    dv = dt * p.tau_mem_inv * ((p.v_leak - state.v) + state.i)
    v_decayed = state.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute threshold updates
    db = dt * p.tau_adapt_inv * (p.v_th - state.b)
    b_decayed = state.b + db

    # compute new spikes
    z_new = threshold(v_decayed - b_decayed, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute b update
    b_new = (1 - z_new) * b_decayed + z_new * state.b
    # compute current jumps
    i_new = i_decayed + input_tensor

    return z_new, LSNNFeedForwardState(v=v_new, i=i_new, b=b_new)



import torch
import numpy as np

from norse.torch.module.leaky_integrator import LICell
from norse.torch.module.lif_correlation import LIFCorrelation
from norse.torch.functional.correlation_sensor import correlation_based_update


def test_lif_correlation_training():
    def generate_random_data(
        seq_length,
        batch_size,
        input_features,
        device="cpu",
        dtype=torch.float,
        dt=0.001,
    ):
        freq = 5
        prob = freq * dt
        mask = torch.rand(
            (seq_length, batch_size, input_features), device=device, dtype=dtype
        )
        x_data = torch.zeros(
            (seq_length, batch_size, input_features),
            device=device,
            dtype=dtype,
            requires_grad=False,
        )
        x_data[mask < prob] = 1.0
        y_data = torch.tensor(1 * (np.random.rand(batch_size) < 0.5), device=device)
        return x_data, y_data

    seq_length = 500
    batch_size = 1
    input_features = 100
    hidden_features = 8
    output_features = 2

    device = "cpu"

    x, y_data = generate_random_data(
        seq_length=seq_length,
        batch_size=batch_size,
        input_features=input_features,
        device=device,
    )

    input_weights = (
        torch.randn((input_features, hidden_features), device=device).float().t()
    )

    recurrent_weights = torch.randn(
        (hidden_features, hidden_features), device=device
    ).float()

    lif_correlation = LIFCorrelation(input_features, hidden_features)
    out = LICell(hidden_features, output_features).to(device)
    log_softmax_fn = torch.nn.LogSoftmax(dim=1)
    loss_fn = torch.nn.NLLLoss()

    linear_update = torch.nn.Linear(2 * 100 * 8, 100 * 8)
    rec_linear_update = torch.nn.Linear(2 * 8 * 8, 8 * 8)

    optimizer = torch.optim.Adam(
        list(linear_update.parameters())
        + [input_weights, recurrent_weights]
        + list(out.parameters()),
        lr=1e-1,
    )

    loss_hist = []
    num_episodes = 3

    for e in range(num_episodes):
        s1 = None
        so = None

        voltages = torch.zeros(seq_length, batch_size, output_features, device=device)
        hidden_voltages = torch.zeros(
            seq_length, batch_size, hidden_features, device=device
        )
        hidden_currents = torch.zeros(
            seq_length, batch_size, hidden_features, device=device
        )

        optimizer.zero_grad()

        for ts in range(seq_length):
            z1, s1 = lif_correlation(
                x[ts, :, :],
                input_weights=input_weights,
                recurrent_weights=recurrent_weights,
                state=s1,
            )

            input_weights = correlation_based_update(
                ts,
                linear_update,
                input_weights.detach(),
                s1.input_correlation_state,
                0.01,
                10,
            )
            recurrent_weights = correlation_based_update(
                ts,
                rec_linear_update,
                recurrent_weights.detach(),
                s1.recurrent_correlation_state,
                0.01,
                10,
            )
            vo, so = out(z1, so)
            hidden_voltages[ts, :, :] = s1.lif_state.v.detach()
            hidden_currents[ts, :, :] = s1.lif_state.i.detach()
            voltages[ts, :, :] = vo

        m, _ = torch.max(voltages, dim=0)

        log_p_y = log_softmax_fn(m)
        loss_val = loss_fn(log_p_y, y_data)

        loss_val.backward()
        optimizer.step()
        loss_hist.append(loss_val.item())
        print(f"{e}/{num_episodes}: {loss_val.item()}")

    # assert loss_hist[0] > loss_hist[1]
    # assert loss_hist[1] > loss_hist[2]
